# -*- coding: utf-8 -*-

import pytest

from sceptre.template_handlers.file import File
from unittest.mock import patch, mock_open


class TestFile(object):

    @pytest.mark.parametrize("project_path,path,output_path", [
        ("my_project_dir",      "my.template.yaml",                               "my_project_dir/templates/my.template.yaml"),        # NOQA
        ("/src/my_project_dir", "my.template.yaml",                               "/src/my_project_dir/templates/my.template.yaml"),   # NOQA
        ("my_project_dir",      "/src/my_project_dir/templates/my.template.yaml", "/src/my_project_dir/templates/my.template.yaml"),   # NOQA
        ("/src/my_project_dir", "/src/my_project_dir/templates/my.template.yaml", "/src/my_project_dir/templates/my.template.yaml"),   # NOQA
    ])
    @patch("builtins.open", new_callable=mock_open, read_data="some_data")
    def test_handler_open(self, mocked_open, project_path, path, output_path):
        template_handler = File(
            name="file_handler",
            arguments={"path": path},
            stack_group_config={"project_path": project_path}
        )
        template_handler.handle()
        mocked_open.assert_called_with(output_path)

    @pytest.mark.parametrize("project_path,path,output_path", [
        ("my_project_dir",      "my.template.yaml.j2",                               "my_project_dir/templates/my.template.yaml.j2"),        # NOQA
        ("/src/my_project_dir", "my.template.yaml.j2",                               "/src/my_project_dir/templates/my.template.yaml.j2"),   # NOQA
        ("my_project_dir",      "/src/my_project_dir/templates/my.template.yaml.j2", "/src/my_project_dir/templates/my.template.yaml.j2"),   # NOQA
        ("/src/my_project_dir", "/src/my_project_dir/templates/my.template.yaml.j2", "/src/my_project_dir/templates/my.template.yaml.j2"),   # NOQA
    ])
    @patch("sceptre.template_handlers.helper.render_jinja_template")
    def test_handler_render(self, mocked_render, project_path, path, output_path):
        template_handler = File(
            name="file_handler",
            arguments={"path": path},
            stack_group_config={"project_path": project_path}
        )
        template_handler.handle()
        mocked_render.assert_called_with(output_path, {"sceptre_user_data": None}, {})

    @pytest.mark.parametrize("project_path,path,output_path", [
        ("my_project_dir",      "my.template.yaml.py",                               "my_project_dir/templates/my.template.yaml.py"),        # NOQA
        ("/src/my_project_dir", "my.template.yaml.py",                               "/src/my_project_dir/templates/my.template.yaml.py"),   # NOQA
        ("my_project_dir",      "/src/my_project_dir/templates/my.template.yaml.py", "/src/my_project_dir/templates/my.template.yaml.py"),   # NOQA
        ("/src/my_project_dir", "/src/my_project_dir/templates/my.template.yaml.py", "/src/my_project_dir/templates/my.template.yaml.py"),   # NOQA
    ])
    @patch("sceptre.template_handlers.helper.call_sceptre_handler")
    def test_handler_handler(self, mocked_handler, project_path, path, output_path):
        template_handler = File(
            name="file_handler",
            arguments={"path": path},
            stack_group_config={"project_path": project_path}
        )
        template_handler.handle()
        mocked_handler.assert_called_with(output_path, None)



# -*- coding: utf-8 -*-

import pytest

from os.path import join, sep

from sceptre.exceptions import PathConversionError
from sceptre.helpers import get_external_stack_name
from sceptre.helpers import normalise_path
from sceptre.helpers import sceptreise_path


class TestHelpers(object):

    def test_get_external_stack_name(self):
        result = get_external_stack_name("prj", "dev/ew1/jump-host")
        assert result == "prj-dev-ew1-jump-host"

    def test_normalise_path_with_valid_path(self):
        path = normalise_path("valid/path")
        assert path == join("valid", "path")

    def test_normalise_path_with_backslashes_in_path(self):
        path = normalise_path("valid\\path")
        assert path == join("valid", "path")

    def test_normalise_path_with_double_backslashes_in_path(self):
        path = normalise_path('valid\\path')
        assert path == join("valid", "path")

    def test_normalise_path_with_leading_slash(self):
        path = normalise_path("/this/is/valid")
        assert path == join("{}this".format(sep), "is", "valid")

    def test_normalise_path_with_leading_backslash(self):
        path = normalise_path('\\this\\path\\is\\valid')
        assert path == join("{}this".format(sep), "path", "is", "valid")

    def test_normalise_path_with_trailing_slash(self):
        with pytest.raises(PathConversionError):
            normalise_path(
                "this/path/is/invalid/"
            )

    def test_normalise_path_with_trailing_backslash(self):
        with pytest.raises(PathConversionError):
            normalise_path(
                'this\\path\\is\\invalid\\'
            )

    def test_sceptreise_path_with_valid_path(self):
        path = 'dev/app/stack'
        assert sceptreise_path(path) == path

    def test_sceptreise_path_with_windows_path(self):
        windows_path = 'dev\\app\\stack'
        assert sceptreise_path(windows_path) == 'dev/app/stack'

    def test_sceptreise_path_with_trailing_slash(self):
        with pytest.raises(PathConversionError):
            sceptreise_path(
                "this/path/is/invalid/"
            )

    def test_sceptreise_path_with_trailing_backslash(self):
        with pytest.raises(PathConversionError):
            sceptreise_path(
                'this\\path\\is\\invalid\\'
            )



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

# TODO modify project metadata (name, author) according to old-docs
#

import os
import sys

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..' + os.path.sep + '..'))
import sceptre  # noqa

# The short X.Y version
version = sceptre.__version__
# The full version, including alpha/beta/rc tags
release = version

project = 'Sceptre'
copyright = '2018, Cloudreach'
author = sceptre.__author__

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx_autodoc_typehints',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'sphinx_click.ext',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set 'language' from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named 'default.css' will overwrite the builtin 'default.css'.
# html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'Sceptredoc'

# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'Sceptre.tex', 'Sceptre Documentation', 'Cloudreach', 'manual')
]

# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, 'sceptre', 'Sceptre Documentation', [author], 1)]

# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        'Sceptre',
        'Sceptre Documentation',
        author,
        'Sceptre',
        'One line description of project.',
        'Miscellaneous',
    )
]

# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project
epub_author = author
epub_publisher = author
epub_copyright = copyright

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']

# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'boto3': (
        'https://boto3.amazonaws.com/v1/documentation/api/latest/',
        'https://boto3.amazonaws.com/v1/documentation/api/latest/objects.inv',
    ),
    'deepdiff': (
        'https://zepworks.com/deepdiff/current/',
        None
    )
}

# other configuration

nitpicky = True

nitpick_ignore = [
    ('py:class', 'json.encoder.JSONEncoder'),
    ('py:class', 'sceptre.config.reader.Attributes'),
    ('py:class', 'sceptre.diffing.stack_differ.DiffType'),
    ('py:obj', 'sceptre.diffing.stack_differ.DiffType'),
    ('py:class', 'DiffType'),
    ('py:class', 'TextIO'),
    ('py:class', '_io.StringIO'),
    ('py:class', 'yaml.loader.SafeLoader'),
    ('py:class', 'yaml.dumper.Dumper'),
    ('py:class',  'cfn_tools.odict.ODict'),
    ('py:class',  'T_Container')
]

set_type_checking_flag = True



# -*- coding: utf-8 -*-

"""
sceptre.template

This module implements a Template class, which stores a CloudFormation template
and implements methods for uploading it to S3.
"""

import logging
import threading
import botocore
import sys

from sceptre.exceptions import TemplateHandlerNotFoundError


class Template(object):
    """
    Template represents an AWS CloudFormation template. It is responsible for
    loading, storing and optionally uploading local templates for use by
    CloudFormation.

    :param name: The name of the template. Should be safe to use in filenames and not contain path segments.
    :type name: str

    :param handler_config: The configuration for a Template handler. Must contain a `type`.
    :type handler_config: dict

    :param sceptre_user_data: A dictionary of arbitrary data to be passed to\
            a handler function in an external Python script.
    :type sceptre_user_data: dict

    :param stack_group_config: The StackGroup config for the Stack.
    :type stack_group_config: dict

    :param connection_manager:
    :type connection_manager: sceptre.connection_manager.ConnectionManager

    :param s3_details:
    :type s3_details: dict
    """

    _boto_s3_lock = threading.Lock()

    def __init__(
        self, name, handler_config, sceptre_user_data,
        stack_group_config, connection_manager=None, s3_details=None
    ):
        self.logger = logging.getLogger(__name__)

        self.name = name
        self.handler_config = handler_config
        if self.handler_config is not None and self.handler_config.get('type') is None:
            self.handler_config['type'] = 'file'
        self.sceptre_user_data = sceptre_user_data
        self.stack_group_config = stack_group_config
        self.connection_manager = connection_manager
        self.s3_details = s3_details

        self._registry = None
        self._body = None

    def __repr__(self):
        return (
            "sceptre.template.Template(name='{0}', handler_config={1}, sceptre_user_data={2}, s3_details={3})".format(
                self.name, self.handler_config, self.sceptre_user_data, self.s3_details
            )
        )

    @property
    def body(self):
        """
        Represents body of the CloudFormation template.

        :returns: The body of the CloudFormation template.
        :rtype: str
        """
        if self._body is None:
            type = self.handler_config.get("type")
            handler_class = self._get_handler_of_type(type)
            handler = handler_class(
                name=self.name,
                arguments={k: v for k, v in self.handler_config.items() if k != "type"},
                sceptre_user_data=self.sceptre_user_data,
                connection_manager=self.connection_manager,
                stack_group_config=self.stack_group_config
            )
            handler.validate()
            body = handler.handle()
            if isinstance(body, bytes):
                body = body.decode('utf-8')
            if not str(body).startswith("---"):
                body = "---\n{}".format(body)
            self._body = body

        return self._body

    def upload_to_s3(self):
        """
        Uploads the template to ``bucket_name`` and returns its URL.

        The Template is uploaded with the ``bucket_key``.

        :returns: The URL of the Template object in S3.
        :rtype: str
        :raises: botocore.exceptions.ClientError

        """
        self.logger.debug("%s - Uploading template to S3...", self.name)

        with self._boto_s3_lock:
            if not self._bucket_exists():
                self._create_bucket()

        # Remove any leading or trailing slashes the user may have added.
        bucket_name = self.s3_details["bucket_name"]
        bucket_key = self.s3_details["bucket_key"]
        bucket_region = self._bucket_region(bucket_name)

        self.logger.debug(
            "%s - Uploading template to: 's3://%s/%s'",
            self.name, bucket_name, bucket_key
        )
        self.connection_manager.call(
            service="s3",
            command="put_object",
            kwargs={
                "Bucket": bucket_name,
                "Key": bucket_key,
                "Body": self.body,
                "ServerSideEncryption": "AES256"
            }
        )

        url = "https://{}.s3.{}.amazonaws.{}/{}".format(
            bucket_name, bucket_region, self._domain_from_region(bucket_region), bucket_key
        )

        self.logger.debug("%s - Template URL: '%s'", self.name, url)

        return url

    def _bucket_exists(self):
        """
        Checks if the bucket ``bucket_name`` exists.

        :returns: Boolean whether the bucket exists
        :rtype: bool
        :raises: botocore.exception.ClientError

        """
        bucket_name = self.s3_details["bucket_name"]
        self.logger.debug(
            "%s - Attempting to find template bucket '%s'",
            self.name, bucket_name
        )
        try:
            self.connection_manager.call(
                service="s3",
                command="head_bucket",
                kwargs={"Bucket": bucket_name}
            )
        except botocore.exceptions.ClientError as exp:
            if exp.response["Error"]["Message"] == "Not Found":
                self.logger.debug(
                    "%s - %s bucket not found.", self.name, bucket_name
                )
                return False
            else:
                raise
        self.logger.debug(
            "%s - Found template bucket '%s'", self.name, bucket_name
        )
        return True

    def _create_bucket(self):
        """
        Create the s3 bucket ``bucket_name``.

        :raises: botocore.exception.ClientError

        """
        bucket_name = self.s3_details["bucket_name"]

        self.logger.debug(
            "%s - Creating new bucket '%s'", self.name, bucket_name
        )

        if self.connection_manager.region == "us-east-1":
            self.connection_manager.call(
                service="s3",
                command="create_bucket",
                kwargs={"Bucket": bucket_name}
            )
        else:
            self.connection_manager.call(
                service="s3",
                command="create_bucket",
                kwargs={
                    "Bucket": bucket_name,
                    "CreateBucketConfiguration": {
                        "LocationConstraint": self.connection_manager.region
                    }
                }
            )

    def get_boto_call_parameter(self):
        """
        Returns the CloudFormation template location.

        Uploads the template to S3 and returns the object's URL, or returns
        the template itself.

        :returns: The boto call parameter for the template.
        :rtype: dict
        """
        # If bucket_name is set to None, it should be ignored and not uploaded.
        if self.s3_details and self.s3_details.get("bucket_name"):
            url = self.upload_to_s3()
            return {"TemplateURL": url}
        else:
            return {"TemplateBody": self.body}

    def _bucket_region(self, bucket_name):
        region = self.connection_manager.call(
            service="s3",
            command="get_bucket_location",
            kwargs={"Bucket": bucket_name}
        ).get("LocationConstraint")
        return region if region else "us-east-1"

    @staticmethod
    def _domain_from_region(region):
        return "com.cn" if region.startswith("cn-") else "com"

    @staticmethod
    def _iterate_entry_points(group, name):
        """
        Helper to determine whether to use pkg_resources or importlib.metadata.
        https://docs.python.org/3/library/importlib.metadata.html
        """
        if sys.version_info < (3, 10):
            from pkg_resources import iter_entry_points
            return iter_entry_points(group, name)
        else:
            from importlib.metadata import entry_points
            return entry_points(group=group, name=name)

    def _get_handler_of_type(self, type):
        """
        Gets a TemplateHandler type from the registry that can be used to get a string
        representation of a CloudFormation template.
        :param type: The type of Template Handler to load
        :type type: str
        :return: Instantiated TemplateHandler
        :rtype: class
        """
        if not self._registry:
            self._registry = {}

            for entry_point in self._iterate_entry_points("sceptre.template_handlers", type):
                self._registry[entry_point.name] = entry_point.load()

        if type not in self._registry:
            raise TemplateHandlerNotFoundError('Handler of type "{0}" not found'.format(type))

        return self._registry[type]



# -*- coding: utf-8 -*-

"""
sceptre.cli

This module implements Sceptre's CLI, and should not be directly imported.
"""

import os

import warnings

import click
import colorama

from sceptre import __version__
from sceptre.cli.new import new_group
from sceptre.cli.create import create_command
from sceptre.cli.prune import prune_command
from sceptre.cli.update import update_command
from sceptre.cli.delete import delete_command
from sceptre.cli.launch import launch_command
from sceptre.cli.diff import diff_command
from sceptre.cli.drift import drift_group
from sceptre.cli.execute import execute_command
from sceptre.cli.describe import describe_group
from sceptre.cli.list import list_group
from sceptre.cli.dump import dump_group
from sceptre.cli.policy import set_policy_command
from sceptre.cli.status import status_command
from sceptre.cli.helpers import catch_exceptions, setup_vars

from sceptre.cli.template import (validate_command,
                                  generate_command,
                                  estimate_cost_command,
                                  fetch_remote_template_command)


@click.group()
@click.version_option(version=__version__, prog_name="Sceptre")
@click.option("--debug", is_flag=True, help="Turn on debug logging.")
@click.option("--dir", "directory", help="Specify sceptre directory.")
@click.option(
    "--output", type=click.Choice(["text", "yaml", "json"]), default="text",
    help="The formatting style for command output.")
@click.option("--no-colour", is_flag=True, help="Turn off output colouring.")
@click.option(
    "--var", multiple=True,
    help="A variable to replace the value of an item in config file.")
@click.option(
    "--var-file", multiple=True, type=click.File("rb"),
    help="A YAML file of variables to replace the values of items in config files.")
@click.option(
    "--ignore-dependencies", is_flag=True,
    help="Ignore dependencies when executing command.")
@click.option(
    "--merge-vars", is_flag=True, default=False,
    help="Merge variables from successive --vars and var files")
@click.pass_context
@catch_exceptions
def cli(
        ctx, debug, directory, output, no_colour, var, var_file, ignore_dependencies, merge_vars
):
    """
    Sceptre is a tool to manage your cloud native infrastructure deployments.
    """
    colorama.init()
    # Enable deprecation warnings
    warnings.simplefilter("always", DeprecationWarning)
    ctx.obj = {
        "user_variables": setup_vars(var_file, var, merge_vars, debug, no_colour),
        "output_format": output,
        "no_colour": no_colour,
        "ignore_dependencies": ignore_dependencies,
        "project_path": directory if directory else os.getcwd()
    }


cli.add_command(new_group)
cli.add_command(create_command)
cli.add_command(update_command)
cli.add_command(delete_command)
cli.add_command(launch_command)
cli.add_command(execute_command)
cli.add_command(validate_command)
cli.add_command(estimate_cost_command)
cli.add_command(generate_command)
cli.add_command(set_policy_command)
cli.add_command(status_command)
cli.add_command(list_group)
cli.add_command(dump_group)
cli.add_command(describe_group)
cli.add_command(fetch_remote_template_command)
cli.add_command(diff_command)
cli.add_command(drift_group)
cli.add_command(prune_command)



# -*- coding: utf-8 -*-

from troposphere import Template

from os import chdir, getcwd


def sceptre_handler(sceptre_user_data):
    t = Template()

    curr_dir = getcwd()
    chdir("..")
    chdir(curr_dir)

    return t.to_json()



import logging
import os
import sys
import traceback

from importlib.machinery import SourceFileLoader
from jinja2 import Environment, select_autoescape, FileSystemLoader, StrictUndefined
from pathlib import Path
from sceptre.exceptions import TemplateSceptreHandlerError, TemplateNotFoundError
from sceptre.config import strategies

logger = logging.getLogger(__name__)

"""
Template handler helpers.
"""


def call_sceptre_handler(path, sceptre_user_data):
    """
    Calls the function `sceptre_handler` within templates that are python
    scripts.

    :param path: A path to the file.
    :type name: str
    :param sceptre_user_data: The sceptre_user_data parameter values.
    :type name: str
    :returns: The string returned from sceptre_handler in the template.
    :rtype: str
    :raises: IOError
    :raises: TemplateSceptreHandlerError
    """
    # Get relative path as list between current working directory and where
    # the template is
    # NB: this is a horrible hack...
    relpath = os.path.relpath(path, os.getcwd()).split(os.path.sep)
    relpaths_to_add = [
        os.path.sep.join(relpath[:i + 1])
        for i in range(len(relpath[:-1]))
    ]
    # Add any directory between the current working directory and where
    # the template is to the python path
    for directory in relpaths_to_add:
        sys.path.append(os.path.join(os.getcwd(), directory))
    logger.debug(
        "Getting CloudFormation from %s", path
    )

    if not os.path.isfile(path):
        raise TemplateNotFoundError("No such template file: '%s'", path)

    module = SourceFileLoader(path, path).load_module()

    try:
        body = module.sceptre_handler(sceptre_user_data)
    except AttributeError as e:
        if 'sceptre_handler' in str(e):
            raise TemplateSceptreHandlerError(
                "The template does not have the required "
                "'sceptre_handler(sceptre_user_data)' function."
            )
        else:
            raise e
    for directory in relpaths_to_add:
        sys.path.remove(os.path.join(os.getcwd(), directory))
    return body


def print_template_traceback(path):
    """
    Prints a stack trace, including only files which are inside a
    'templates' directory. The function is intended to give the operator
    instant feedback about why their templates are failing to compile.

    :param path: A path to the file.
    :type name: str
    :rtype: None
    """

    def _print_frame(filename, line, fcn, line_text):
        logger.error("{}:{}:  Template error in '{}'\n=> `{}`".format(
            filename, line, fcn, line_text))

    try:
        _, _, tb = sys.exc_info()
        stack_trace = traceback.extract_tb(tb)
        search_string = os.path.join('', 'templates', '')
        if search_string in path:
            template_path = path.split(search_string)[0] + search_string
        else:
            return
        for frame in stack_trace:
            if isinstance(frame, tuple):
                # Python 2 / Old style stack frame
                if template_path in frame[0]:
                    _print_frame(frame[0], frame[1], frame[2], frame[3])
            else:
                if template_path in frame.filename:
                    _print_frame(frame.filename, frame.lineno, frame.name, frame.line)
    except Exception as tb_exception:
        logger.error(
            'A template error occured. ' +
            'Additionally, a traceback exception occured. Exception: %s',
            tb_exception
        )


def render_jinja_template(path, jinja_vars, j2_environment):
    """
    Renders a jinja template.

    Sceptre supports passing sceptre_user_data to JSON and YAML
    CloudFormation templates using Jinja2 templating.

    :param path: The path to the template file.
    :type path: str
    :param jinja_vars: Dict of variables to render into the template.
    :type jinja_vars: dict
    :param j2_environment: The jinja2 environment.
    :type stack_group_config: dict

    :returns: The body of the CloudFormation template.
    :rtype: str
    """
    path = Path(path)
    if not path.exists():
        raise TemplateNotFoundError("No such template file: '%s'", path)

    logger.debug("%s Rendering CloudFormation template", path)
    default_j2_environment_config = {
        "autoescape": select_autoescape(
            disabled_extensions=('j2',),
            default=True,
        ),
        "loader": FileSystemLoader(path.parent),
        "undefined": StrictUndefined
    }
    j2_environment_config = strategies.dict_merge(
        default_j2_environment_config, j2_environment
    )
    j2_environment = Environment(**j2_environment_config)

    template = j2_environment.get_template(path.name)

    body = template.render(**jinja_vars)
    return body



# -*- coding: utf-8 -*-

import errno
import os
from unittest.mock import patch, sentinel, MagicMock

import pytest
import yaml
from click.testing import CliRunner
from freezegun import freeze_time

from sceptre.config.reader import ConfigReader
from sceptre.context import SceptreContext
from sceptre.exceptions import (
    DependencyDoesNotExistError,
    VersionIncompatibleError,
    ConfigFileNotFoundError,
    InvalidSceptreDirectoryError,
    InvalidConfigFileError
)


class TestConfigReader(object):
    @patch("sceptre.config.reader.ConfigReader._check_valid_project_path")
    def setup_method(self, test_method, mock_check_valid_project_path):
        self.runner = CliRunner()
        self.test_project_path = os.path.join(
            os.getcwd(), "tests", "fixtures"
        )
        self.context = SceptreContext(
            project_path=self.test_project_path,
            command_path="A"
        )

    def test_config_reader_correctly_initialised(self):
        config_reader = ConfigReader(self.context)
        assert config_reader.context == self.context

    def test_config_reader_with_invalid_path(self):
        with pytest.raises(InvalidSceptreDirectoryError):
            ConfigReader(SceptreContext("/path/does/not/exist", "example"))

    def create_project(self):
        """
        Creates a new random temporary directory with a config subdirectory
        """
        with self.runner.isolated_filesystem():
            project_path = os.path.abspath('./example')
        config_dir = os.path.join(project_path, "config")
        os.makedirs(config_dir)
        return (project_path, config_dir)

    def write_config(self, abs_path, config):
        """
        Writes a configuration dict to the specified path as YAML
        """
        if abs_path.endswith(".yaml"):
            dir_path = os.path.split(abs_path)[0]
        if not os.path.exists(dir_path):
            try:
                os.makedirs(dir_path)
            except OSError as exc:
                if exc.errno != errno.EEXIST:
                    raise

        with open(abs_path, 'w') as config_file:
            yaml.safe_dump(
                config, stream=config_file, default_flow_style=False
            )

    @pytest.mark.parametrize("filepaths,target", [
        (
            ["A/1.yaml"], "A/1.yaml"
        ),
        (
            ["A/1.yaml", "A/B/1.yaml"], "A/B/1.yaml"
        ),
        (
            ["A/1.yaml", "A/B/1.yaml", "A/B/C/1.yaml"], "A/B/C/1.yaml"
        )
    ])
    def test_read_reads_config_file(self, filepaths, target):
        project_path, config_dir = self.create_project()

        for rel_path in filepaths:
            config = {"filepath": rel_path}
            abs_path = os.path.join(config_dir, rel_path)
            self.write_config(abs_path, config)

        self.context.project_path = project_path
        config = ConfigReader(self.context).read(target)

        assert config == {
            "project_path": project_path,
            "stack_group_path": os.path.split(target)[0],
            "filepath": target
        }

    def test_read_nested_configs(self):
        with self.runner.isolated_filesystem():
            project_path = os.path.abspath('./example')
            config_dir = os.path.join(project_path, "config")
            stack_group_dir_a = os.path.join(config_dir, "A")
            stack_group_dir_b = os.path.join(stack_group_dir_a, "B")
            stack_group_dir_c = os.path.join(stack_group_dir_b, "C")

            os.makedirs(stack_group_dir_c)
            config_filename = "config.yaml"

            config_a = {"keyA": "A", "shared": "A"}
            with open(os.path.join(stack_group_dir_a, config_filename), 'w') as\
                    config_file:
                yaml.safe_dump(
                    config_a, stream=config_file, default_flow_style=False
                )

            config_b = {"keyB": "B", "parent": "{{ keyA }}", "shared": "B"}
            with open(os.path.join(stack_group_dir_b, config_filename), 'w') as\
                    config_file:
                yaml.safe_dump(
                    config_b, stream=config_file, default_flow_style=False
                )

            config_c = {"keyC": "C", "parent": "{{ keyB }}", "shared": "C"}
            with open(os.path.join(stack_group_dir_c, config_filename), 'w') as\
                    config_file:
                yaml.safe_dump(
                    config_c, stream=config_file, default_flow_style=False
                )

            self.context.project_path = project_path
            reader = ConfigReader(self.context)

            config_a = reader.read("A/config.yaml")

            assert config_a == {
                "project_path": project_path,
                "stack_group_path": "A",
                "keyA": "A",
                "shared": "A"
            }

            config_b = reader.read("A/B/config.yaml")

            assert config_b == {
                "project_path": project_path,
                "stack_group_path": "A/B",
                "keyA": "A",
                "keyB": "B",
                "shared": "B",
                "parent": "A"
            }

            config_c = reader.read(
                "A/B/C/config.yaml"
            )

            assert config_c == {
                "project_path": project_path,
                "stack_group_path": "A/B/C",
                "keyA": "A",
                "keyB": "B",
                "keyC": "C",
                "shared": "C",
                "parent": "B"
            }

    def test_read_reads_config_file_with_base_config(self):
        with self.runner.isolated_filesystem():
            project_path = os.path.abspath('./example')
            config_dir = os.path.join(project_path, "config")
            stack_group_dir = os.path.join(config_dir, "A")

            os.makedirs(stack_group_dir)

            config = {"config": "config"}
            with open(os.path.join(stack_group_dir, "stack.yaml"), 'w') as\
                    config_file:
                yaml.safe_dump(
                    config, stream=config_file, default_flow_style=False
                )

            base_config = {
                "base_config": "base_config"
            }
            self.context.project_path = project_path
            config = ConfigReader(self.context).read(
                "A/stack.yaml", base_config
            )

            assert config == {
                "project_path": project_path,
                "stack_group_path": "A",
                "config": "config",
                "base_config": "base_config"
            }

    def test_read_with_nonexistant_filepath(self):
        project_path, config_dir = self.create_project()
        self.context.project_path = project_path
        with pytest.raises(ConfigFileNotFoundError):
            ConfigReader(self.context).read("stack.yaml")

    def test_read_with_empty_config_file(self):
        config_reader = ConfigReader(self.context)
        config = config_reader.read(
            "account/stack-group/region/subnets.yaml"
        )
        assert config == {
            "project_path": self.test_project_path,
            "stack_group_path": "account/stack-group/region"
        }

    def test_read_with_templated_config_file(self):
        self.context.user_variables = {"variable_key": "user_variable_value"}
        config_reader = ConfigReader(self.context)

        config_reader.templating_vars["stack_group_config"] = {
            "region": "region_region",
            "project_code": "account_project_code",
            "required_version": "'>1.0'",
            "template_bucket_name": "stack_group_template_bucket_name"
        }
        os.environ["TEST_ENV_VAR"] = "environment_variable_value"
        config = config_reader.read(
            "account/stack-group/region/security_groups.yaml"
        )

        assert config == {
            'project_path': self.context.project_path,
            "stack_group_path": "account/stack-group/region",
            "parameters": {
                "param1": "user_variable_value",
                "param2": "environment_variable_value",
                "param3": "region_region",
                "param4": "account_project_code",
                "param5": ">1.0",
                "param6": "stack_group_template_bucket_name"
            }
        }

    def test_aborts_on_incompatible_version_requirement(self):
        config = {
            'required_version': '<0'
        }
        with pytest.raises(VersionIncompatibleError):
            ConfigReader(self.context)._check_version(config)

    @freeze_time("2012-01-01")
    @pytest.mark.parametrize("stack_name,config,expected", [
        (
            "name",
            {
                "template_bucket_name": "bucket-name",
                "template_key_prefix": "prefix",
                "region": "eu-west-1"
            },
            {
                "bucket_name": "bucket-name",
                "bucket_key": "prefix/name/2012-01-01-00-00-00-000000Z.json"
            }
        ),
        (
            "name",
            {
                "template_bucket_name": "bucket-name",
                "region": "eu-west-1"
            },
            {
                "bucket_name": "bucket-name",
                "bucket_key": "name/2012-01-01-00-00-00-000000Z.json"
            }
        ),
        (
            "name",
            {
                "template_bucket_name": "bucket-name",
            },
            {
                "bucket_name": "bucket-name",
                "bucket_key": "name/2012-01-01-00-00-00-000000Z.json"
            }
        ),
        (
            "name", {}, None
        )
    ]
    )
    def test_collect_s3_details(self, stack_name, config, expected):
        details = ConfigReader._collect_s3_details(stack_name, config)
        assert details == expected

    @patch("sceptre.config.reader.ConfigReader._collect_s3_details")
    @patch("sceptre.config.reader.Stack")
    def test_construct_stacks_constructs_stack(
        self, mock_Stack, mock_collect_s3_details
    ):
        mock_Stack.return_value = sentinel.stack
        sentinel.stack.dependencies = []

        mock_collect_s3_details.return_value = sentinel.s3_details
        self.context.project_path = os.path.abspath("tests/fixtures-vpc")
        self.context.command_path = "account/stack-group/region/vpc.yaml"
        stacks = ConfigReader(self.context).construct_stacks()
        mock_Stack.assert_any_call(
            name="account/stack-group/region/vpc",
            project_code="account_project_code",
            template_path="path/to/template",
            template_handler_config=None,
            region="region_region",
            profile="account_profile",
            parameters={"param1": "val1"},
            sceptre_user_data={},
            hooks={},
            s3_details=sentinel.s3_details,
            dependencies=["child/level", "top/level"],
            iam_role=None,
            iam_role_session_duration=None,
            role_arn=None,
            protected=False,
            tags={},
            external_name=None,
            notifications=None,
            on_failure=None,
            stack_timeout=0,
            required_version='>1.0',
            template_bucket_name='stack_group_template_bucket_name',
            template_key_prefix=None,
            ignore=False,
            obsolete=False,
            stack_group_config={
                "project_path": self.context.project_path,
                "custom_key": "custom_value"
            }
        )

        assert stacks == ({sentinel.stack}, {sentinel.stack})

    @pytest.mark.parametrize("command_path,filepaths,expected_stacks,expected_command_stacks,full_scan", [
        (
            "",
            ["A/1.yaml"],
            {"A/1"},
            {"A/1"},
            False
        ),
        (
            "",
            ["A/1.yaml", "A/2.yaml", "A/3.yaml"],
            {"A/3", "A/2", "A/1"},
            {"A/3", "A/2", "A/1"},
            False
        ),
        (
            "",
            ["A/1.yaml", "A/A/1.yaml"],
            {"A/1", "A/A/1"},
            {"A/1", "A/A/1"},
            False
        ),
        (
            "",
            ["A/1.yaml", "A/A/1.yaml", "A/A/2.yaml"],
            {"A/1", "A/A/1", "A/A/2"},
            {"A/1", "A/A/1", "A/A/2"},
            False
        ),
        (
            "",
            ["A/A/1.yaml", "A/B/1.yaml"],
            {"A/A/1", "A/B/1"},
            {"A/A/1", "A/B/1"},
            False
        ),
        (
            "Abd",
            ["Abc/1.yaml", "Abd/1.yaml"],
            {"Abd/1"},
            {"Abd/1"},
            False
        ),
        (
            "Abd",
            ["Abc/1.yaml", "Abd/Abc/1.yaml", "Abd/2.yaml"],
            {"Abd/2", "Abd/Abc/1"},
            {"Abd/2", "Abd/Abc/1"},
            False
        ),
        (
            "Abd/Abc",
            ["Abc/1.yaml", "Abd/Abc/1.yaml", "Abd/2.yaml"],
            {"Abd/Abc/1"},
            {"Abd/Abc/1"},
            False
        ),
        (
            "Ab",
            ["Abc/1.yaml", "Abd/1.yaml"],
            set(),
            set(),
            False
        ),
        (
            "Abd/Abc",
            ["Abc/1.yaml", "Abd/Abc/1.yaml", "Abd/2.yaml"],
            {"Abc/1", "Abd/Abc/1", "Abd/2"},
            {"Abd/Abc/1"},
            True
        ),
    ])
    def test_construct_stacks_with_valid_config(
        self, command_path, filepaths, expected_stacks, expected_command_stacks, full_scan
    ):
        project_path, config_dir = self.create_project()

        for rel_path in filepaths:

            config = {
                "region": "region",
                "project_code": "project_code",
                "template_path": rel_path
            }

            abs_path = os.path.join(config_dir, rel_path)
            self.write_config(abs_path, config)

        self.context.project_path = project_path
        self.context.command_path = command_path
        self.context.full_scan = full_scan
        config_reader = ConfigReader(self.context)
        all_stacks, command_stacks = config_reader.construct_stacks()
        assert {str(stack) for stack in all_stacks} == expected_stacks
        assert {str(stack) for stack in command_stacks} == expected_command_stacks

    @pytest.mark.parametrize("filepaths, del_key", [
        (["A/1.yaml"], "project_code"),
        (["A/1.yaml"], "region"),
    ])
    def test_missing_attr(
        self, filepaths, del_key
    ):
        project_path, config_dir = self.create_project()

        for rel_path in filepaths:

            config = {
                "project_code": "project_code",
                "region": "region",
                "template_path": rel_path
            }
            # Delete the mandatory key to be tested.
            del config[del_key]

            abs_path = os.path.join(config_dir, rel_path)
            self.write_config(abs_path, config)

        self.context.project_path = project_path
        try:
            config_reader = ConfigReader(self.context)
            all_stacks, command_stacks = config_reader.construct_stacks()
        except InvalidConfigFileError as e:
            # Test that the missing key is reported.
            assert del_key in str(e)
        except Exception:
            raise
        else:
            assert False

    @pytest.mark.parametrize("filepaths, dependency", [
        (["A/1.yaml", "B/1.yaml", "B/2.yaml"], "A/1.yaml"),
        (["A/1.yaml", "B/1.yaml", "B/2.yaml"], "B/1.yaml"),
    ])
    def test_existing_dependency(
        self, filepaths, dependency
    ):
        project_path, config_dir = self.create_project()

        for rel_path in filepaths:
            # Set up config with reference to an existing stack
            config = {
                "project_code": "project_code",
                "region": "region",
                "template_path": rel_path,
                "dependencies": [dependency]
            }

            abs_path = os.path.join(config_dir, rel_path)
            self.write_config(abs_path, config)

        self.context.project_path = project_path
        try:
            config_reader = ConfigReader(self.context)
            all_stacks, command_stacks = config_reader.construct_stacks()
        except Exception:
            raise
        else:
            assert True

    @pytest.mark.parametrize("filepaths, dependency", [
        (["A/1.yaml", "B/1.yaml", "B/2.yaml"], "A/2.yaml"),
        (["A/1.yaml", "B/1.yaml", "B/2.yaml"], "1.yaml"),
    ])
    def test_missing_dependency(
        self, filepaths, dependency
    ):
        project_path, config_dir = self.create_project()

        for rel_path in filepaths:
            # Set up config with reference to non-existing stack
            config = {
                "project_code": "project_code",
                "region": "region",
                "template_path": rel_path,
                "dependencies": [dependency]
            }

            abs_path = os.path.join(config_dir, rel_path)
            self.write_config(abs_path, config)

        self.context.project_path = project_path
        try:
            config_reader = ConfigReader(self.context)
            all_stacks, command_stacks = config_reader.construct_stacks()
        except DependencyDoesNotExistError as e:
            # Test that the missing dependency is reported.
            assert dependency in str(e)
        except Exception:
            raise
        else:
            assert False

    @pytest.mark.parametrize(
        "filepaths, dependency, parent_config_path", [
            (["A/1.yaml", "A/2.yaml", "B/1.yaml"], "B/1.yaml", 'A/config.yaml'),
            (["A/1.yaml", "A/2.yaml"], "A/1.yaml", 'A/config.yaml'),
        ]
    )
    def test_inherited_dependency_already_resolved(self, filepaths, dependency, parent_config_path):
        project_path, config_dir = self.create_project()
        parent_config = {
            'dependencies': [dependency]
        }
        abs_path = os.path.join(config_dir, parent_config_path)
        self.write_config(abs_path, parent_config)

        for rel_path in filepaths:
            # Set up config with reference to an existing stack
            config = {
                "project_code": "project_code",
                "region": "region",
                "template_path": rel_path,
            }

            abs_path = os.path.join(config_dir, rel_path)
            self.write_config(abs_path, config)
        self.context.project_path = project_path
        try:
            config_reader = ConfigReader(self.context)
            all_stacks, command_stacks = config_reader.construct_stacks()
        except Exception:
            raise
        else:
            assert True

    def test_resolve_node_tag(self):
        mock_loader = MagicMock(yaml.Loader)
        mock_loader.resolve.return_value = "new_tag"

        mock_node = MagicMock(yaml.Node)
        mock_node.tag = "old_tag"
        mock_node.value = "String"

        config_reader = ConfigReader(self.context)
        new_node = config_reader.resolve_node_tag(mock_loader, mock_node)

        assert new_node.tag == 'new_tag'



from sceptre.resolvers import Resolver


class CustomResolver(Resolver):
    def __init__(self, *args, **kwargs):
        super(CustomResolver, self).__init__(*args, **kwargs)

    def resolve(self):
        return "value"



# -*- coding: utf-8 -*-

import tempfile
import pytest

from sceptre.resolvers.file_contents import FileContents


class TestFileContentsResolver(object):

    def setup_method(self, test_method):
        self.file_contents_resolver = FileContents(
            argument=None
        )

    def test_resolving_with_existing_file(self):
        with tempfile.NamedTemporaryFile(mode='w+') as f:
            f.write("file contents")
            f.seek(0)
            self.file_contents_resolver.argument = f.name
            result = self.file_contents_resolver.resolve()

        assert result == "file contents"

    def test_resolving_with_non_existant_file(self):
        with pytest.raises(IOError):
            self.file_contents_resolver.argument = "/non_existant_file"
            self.file_contents_resolver.resolve()

    def test_resolving_with_file_path_non_string_type(self):
        with pytest.raises(TypeError):
            self.file_contents_resolver.argument = None
            self.file_contents_resolver.resolve()



import click
from colorama import Fore, Style

from sceptre.cli.helpers import catch_exceptions, stack_status_exit_code
from sceptre.context import SceptreContext
from sceptre.exceptions import CannotPruneStackError
from sceptre.plan.plan import SceptrePlan
from sceptre.stack import Stack

PATH_FOR_WHOLE_PROJECT = "."


@click.command(name="prune", short_help="Deletes all obsolete stacks in the project")
@click.option(
    "-y", "--yes", is_flag=True, help="Assume yes to all questions."
)
@click.argument("path", default=PATH_FOR_WHOLE_PROJECT)
@click.pass_context
@catch_exceptions
def prune_command(ctx, yes: bool, path):
    """
    This command deletes all obsolete stacks in the project. Only obsolete stacks can be deleted
    via prune; If any non-obsolete stacks depend on obsolete stacks, an error will be
    raised and this command will fail.
    """
    context = SceptreContext(
        command_path=path,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        options=ctx.obj.get("options"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies"),
        full_scan=True,
    )
    pruner = Pruner(context)
    pruner.print_operations()
    if not yes and pruner.prune_count > 0:
        pruner.confirm()

    code = pruner.prune()
    exit(code)


class Pruner:
    """Pruner is a utility to coordinate the flow of deleting all stacks in the project that
    are marked "obsolete".

    Note: The command_path on the passed context will be ignored; This command operates on the
    entire project rather than on any particular command path.

    :param context: The Sceptre context to use for pruning
    :param plan_factory: A callable with the signature of (SceptreContext) -> SceptrePlan
    """
    def __init__(self, context:  SceptreContext, plan_factory=SceptrePlan):
        self._context = context
        self._make_plan = plan_factory

        self._plan = None

    def confirm(self):
        self._confirm_prune()

    def print_operations(self):
        plan = self._create_plan()

        if not self._plan_has_obsolete_stacks(plan):
            self._print_no_obsolete_stacks()
            return

        self._print_stacks_to_be_deleted(plan)

    @property
    def prune_count(self) -> 0:
        plan = self._create_plan()
        if self._plan_has_obsolete_stacks(plan):
            return len(list(plan))
        return 0

    def prune(self) -> int:
        plan = self._create_plan()

        if not self._plan_has_obsolete_stacks(plan):
            return 0

        if not self._context.ignore_dependencies:
            self._validate_plan_for_dependencies_on_obsolete_stacks(plan)

        code = self._prune(plan)
        return code

    def _create_plan(self):
        if not self._plan:
            context = self._context.clone()
            context.full_scan = True
            plan = self._make_plan(self._context)
            if context.command_path == PATH_FOR_WHOLE_PROJECT:
                stacks = plan.graph
            else:
                stacks = plan.command_stacks

            plan.command_stacks = {
                stack for stack in stacks if stack.obsolete
            }
            self._resolve_plan(plan)
            self._plan = plan
        return self._plan

    def _plan_has_obsolete_stacks(self, plan: SceptrePlan):
        return len(plan.command_stacks) > 0

    def _print_no_obsolete_stacks(self):
        click.echo("* There are no stacks marked obsolete, so there is nothing to prune.")

    def _resolve_plan(self, plan: SceptrePlan):
        if len(plan.command_stacks) > 0:
            # Prune is actually a particular kind of filtered deletion, so we use delete as the actual
            # resolved command.
            plan.resolve(plan.delete.__name__, reverse=True)

    def _validate_plan_for_dependencies_on_obsolete_stacks(self, plan: SceptrePlan):
        def check_for_non_obsolete_dependencies(stack: Stack):
            # If we've already established it as an obsolete stack to delete, we're good.
            if stack in plan.command_stacks:
                return

            # This check shouldn't be necessary, but we're just double-checking that it is indeed
            # not obsolete.
            if stack.obsolete:
                return

            # Theoretically, we've already gathered up ALL obsolete stacks as command stacks. If
            # we've hit this line, there's a problem. Now we just need to know what caused it. This
            # block climbs down the dependency graph to see which obsolete stack caused this stack
            # to be included in the plan.
            for dependency in stack.dependencies:
                if dependency.obsolete:
                    raise CannotPruneStackError(
                        f"Cannot prune obsolete stack {dependency.name} because stack {stack.name} "
                        f"depends on it but is not obsolete."
                    )

            # If we get to this point, it means this stack isn't obsolete and none of its dependencies
            # are either. That only happens it depends on another non-obsolete stack that depends on
            # an obsolete stack. As a result, we're not going to blow up here and instead will
            # continue iterating on the plan and will raise the error on a stack that directly
            # depends on the obsolete stack.
            return

        for stack in plan:
            check_for_non_obsolete_dependencies(stack)

    def _print_stacks_to_be_deleted(self, plan: SceptrePlan):
        delete_msg = "* The following obsolete stacks will be deleted (if they exist on AWS):\n"

        stacks_list = ''
        for stack in plan:
            # It's possible there could be stacks in the plan that aren't obsolete because those
            # stacks depend on obsolete stacks. They won't pass validation, but that's not the
            # point of this method. We'll just skip those here and fail validation later.
            if not stack.obsolete:
                continue
            stacks_list += "{}{}{}\n".format(Fore.YELLOW, stack.name, Style.RESET_ALL)

        click.echo(delete_msg + stacks_list)

    def _confirm_prune(self):
        click.confirm("Do you want to delete these stacks?", abort=True)

    def _prune(self, plan: SceptrePlan):
        responses = plan.delete()
        return stack_status_exit_code(responses.values())



# -*- coding: utf-8 -*-
from contextlib import contextmanager
from os import sep

from sceptre.exceptions import PathConversionError


def get_external_stack_name(project_code, stack_name):
    """
    Returns the name given to a stack in CloudFormation.
    :param project_code: The project code, as defined in config.yaml.
    :type project_code: str
    :param stack_name: The name of the stack.
    :type stack_name: str
    :returns: The name given to the stack in CloudFormation.
    :rtype: str
    """
    return "-".join([
        project_code,
        stack_name.replace("/", "-")
    ])


def mask_key(key):
    """
    Returns an masked version of ``key``.

    Returned version has all but the last four characters are replaced with the
    character "*".

    :param key: The string to mask.
    :type key: str
    :returns: An masked version of the key
    :rtype: str
    """
    num_mask_chars = len(key) - 4

    return "".join([
        "*" if i < num_mask_chars else c
        for i, c in enumerate(key)
    ])


def _call_func_on_values(func, attr, cls):
    """
    Searches through dictionary or list for objects of type `cls` and calls the
    supplied function `func`. Supports nested dictionaries and lists.
    Does not detect objects used as keys in dictionaries.

    :param attr: A dictionary or list to search through.
    :type attr: dict or list
    :return: The dictionary or list structure.
    :rtype: dict or list
    """

    def func_on_instance(key):
        if isinstance(value, cls):
            func(attr, key, value)
        elif isinstance(value, list) or isinstance(value, dict):
            _call_func_on_values(func, value, cls)

    if isinstance(attr, dict):
        for key, value in attr.items():
            func_on_instance(key)
    elif isinstance(attr, list):
        for index, value in enumerate(attr):
            func_on_instance(index)
    return attr


def normalise_path(path):
    """
    Converts a path to use correct path separator.
    Raises an PathConversionError if the path has a
    trailing slash.
    :param path: A directory path
    :type path: str
    :raises: sceptre.exceptions.PathConversionError
    :returns: A normalised path with forward slashes.
    :returns: string
    """
    if sep == '/':
        path = path.replace('\\', '/')
    elif sep == '\\':
        path = path.replace('/', '\\')
    if path.endswith("/") or path.endswith("\\"):
        raise PathConversionError(
            "'{0}' is an invalid path string. Paths should "
            "not have trailing slashes.".format(path)
        )
    return path


def sceptreise_path(path):
    """
    Converts a path to use correct sceptre path separator.
    Raises an PathConversionError if the path has a
    trailing slash.
    :param path: A directory path
    :type path: str
    :raises: sceptre.exceptions.PathConversionError
    :returns: A normalised path with forward slashes.
    :returns: string
    """
    path = path.replace('\\', '/')
    if path.endswith("/") or path.endswith("\\"):
        raise PathConversionError(
            "'{0}' is an invalid path string. Paths should "
            "not have trailing slashes.".format(path)
        )
    return path


@contextmanager
def null_context():
    """A context manager that does nothing. This is identical to the nullcontext in py3.7+, but isn't
    available in py3.6, so providing it here instead.
    """
    yield



# -*- coding: utf-8 -*-

"""
sceptre.config.strategies

This module contains the implementations of the strategies used to merge config
attributes.
"""
from copy import deepcopy


def list_join(a, b):
    """
    Takes two lists and joins them.

    :param a: A list.
    :type a: list
    :param b: A list.
    :type b: list
    :returns: A joined list from the two parameters.
    :rtype: list
    """
    if a and not isinstance(a, list):
        raise TypeError('{} is not a list'.format(a))

    if b and not isinstance(b, list):
        raise TypeError('{} is not a list'.format(b))

    if a is None:
        return deepcopy(b)

    if b is not None:
        return deepcopy(a + b)

    return deepcopy(a)


def dict_merge(a, b):
    """
    Takes two dictionaries and merges them.

    :param a: A dictionary.
    :type a: dict
    :param b: A dictionary.
    :type b: dict
    :returns: A merged dict.
    :rtype: dict
    """
    if a and not isinstance(a, dict):
        raise TypeError('{} is not a dict'.format(a))
    if b and not isinstance(b, dict):
        raise TypeError('{} is not a dict'.format(b))

    if a is None:
        return deepcopy(b)

    if b is not None:
        return deepcopy({**a, **b})

    return deepcopy(a)


def child_wins(a, b):
    """
    Always returns the second parameter.

    :param a: An object.
    :type a: object
    :param b: An object.
    :type b: object
    :returns: b
    """
    return b



import difflib
import json
from collections import defaultdict
from copy import deepcopy
from typing import Union, Optional
from unittest.mock import Mock, PropertyMock

import cfn_flip
import pytest
import yaml

from sceptre.diffing.stack_differ import (
    StackDiffer,
    StackConfiguration,
    DiffType,
    DeepDiffStackDiffer,
    DifflibStackDiffer
)
from sceptre.plan.actions import StackActions
from sceptre.stack import Stack


class ImplementedStackDiffer(StackDiffer):

    def __init__(self, command_capturer: Mock):
        super().__init__()
        self.command_capturer = command_capturer

    def compare_templates(self, deployed: str, generated: str) -> DiffType:
        return self.command_capturer.compare_templates(deployed, generated)

    def compare_stack_configurations(
        self,
        deployed: Optional[StackConfiguration],
        generated: StackConfiguration
    ) -> DiffType:
        return self.command_capturer.compare_stack_configurations(deployed, generated)


class TestStackDiffer:

    def setup_method(self, method):
        self.name = 'my/stack'
        self.external_name = "full-stack-name"
        self.role_arn = 'role_arn'
        self.parameters_on_stack_config = {
            'param': 'some_value'
        }
        self.tags = {
            'tag_name': 'tag_value'
        }
        self.notifications = [
            'notification_arn1'
        ]
        self.sceptre_user_data = {}

        self.deployed_parameters = deepcopy(self.parameters_on_stack_config)
        self.deployed_parameter_defaults = {}
        self.deployed_no_echo_parameters = []
        self.deployed_parameter_types = defaultdict(lambda: 'String')
        self.local_no_echo_parameters = []
        self.deployed_tags = dict(self.tags)
        self.deployed_notification_arns = list(self.notifications)
        self.deployed_role_arn = self.role_arn

        self.command_capturer = Mock()
        self.differ = ImplementedStackDiffer(self.command_capturer)
        self.stack_status = 'CREATE_COMPLETE'

        self._stack = None
        self._actions = None
        self._parameters = None

    @property
    def parameters_on_stack(self):
        if self._parameters is None:
            self._parameters = deepcopy(self.parameters_on_stack_config)
        return self._parameters

    @property
    def stack(self) -> Union[Stack, Mock]:
        if not self._stack:
            self._stack = Mock(
                spec=Stack,
                external_name=self.external_name,
                _parameters=self.parameters_on_stack,
                role_arn=self.role_arn,
                tags=self.tags,
                notifications=self.notifications,
                __sceptre_user_data=self.sceptre_user_data
            )
            self._stack.name = self.name
            type(self._stack).parameters = PropertyMock(side_effect=lambda: self.parameters_on_stack)
        return self._stack

    @property
    def actions(self) -> Union[StackActions, Mock]:
        if not self._actions:
            self._actions = Mock(
                **{
                    'spec': StackActions,
                    'stack': self.stack,
                    'describe.side_effect': self.describe_stack,
                    'fetch_remote_template_summary.side_effect': self.get_remote_template_summary,
                    'fetch_local_template_summary.side_effect': self.get_local_template_summary,
                }
            )
        return self._actions

    def describe_stack(self):
        return {
            'Stacks': [
                {
                    'StackName': self.stack.external_name,
                    'Parameters': [
                        {
                            'ParameterKey': key,
                            'ParameterValue': value,
                            'ResolvedValue': "I'm resolved and don't matter for the diff!"
                        }
                        for key, value in self.deployed_parameters.items()
                    ],
                    'StackStatus': self.stack_status,
                    'NotificationARNs': self.deployed_notification_arns,
                    'RoleARN': self.deployed_role_arn,
                    'Tags': [
                        {
                            'Key': key,
                            'Value': value
                        }
                        for key, value in self.deployed_tags.items()
                    ],
                },
            ],
        }

    def get_remote_template_summary(self):
        params = []
        for param, value in self.deployed_parameters.items():
            entry = {
                'ParameterKey': param,
                'ParameterType': self.deployed_parameter_types[param]
            }
            if param in self.deployed_parameter_defaults:
                default_value = self.deployed_parameter_defaults[param]
                if 'List' in entry['ParameterType']:
                    default_value = ', '.join(val.strip() for val in default_value.split(','))
                entry['DefaultValue'] = default_value
            if param in self.deployed_no_echo_parameters:
                entry['NoEcho'] = True

            params.append(entry)

        return {
            'Parameters': params
        }

    def get_local_template_summary(self):
        params = []
        for param, value in self.parameters_on_stack.items():
            entry = {
                'ParameterKey': param
            }
            if param in self.local_no_echo_parameters:
                entry['NoEcho'] = True
            params.append(entry)

        return {
            'Parameters': params
        }

    @property
    def expected_generated_config(self):
        return StackConfiguration(
            stack_name=self.external_name,
            parameters=self.parameters_on_stack_config,
            stack_tags=deepcopy(self.tags),
            notifications=deepcopy(self.notifications),
            role_arn=self.role_arn
        )

    @property
    def expected_deployed_config(self):
        return StackConfiguration(
            stack_name=self.external_name,
            parameters=self.deployed_parameters,
            stack_tags=deepcopy(self.deployed_tags),
            notifications=deepcopy(self.deployed_notification_arns),
            role_arn=self.deployed_role_arn
        )

    def test_diff__compares_deployed_template_to_generated_template(self):
        self.differ.diff(self.actions)

        self.command_capturer.compare_templates.assert_called_with(
            self.actions.fetch_remote_template.return_value,
            self.actions.generate.return_value
        )

    def test_diff__template_diff_is_value_returned_by_implemented_differ(self):
        diff = self.differ.diff(self.actions)

        assert diff.template_diff == self.command_capturer.compare_templates.return_value

    def test_diff__compares_deployed_stack_config_to_generated_stack_config(self):
        self.deployed_parameters['new'] = 'value'

        self.differ.diff(self.actions)

        self.command_capturer.compare_stack_configurations.assert_called_with(
            self.expected_deployed_config,
            self.expected_generated_config
        )

    def test_diff__config_diff_is_value_returned_by_implemented_differ(self):
        diff = self.differ.diff(self.actions)

        assert diff.config_diff == self.command_capturer.compare_stack_configurations.return_value

    def test_diff__returned_diff_has_stack_name_of_external_name(self):
        diff = self.differ.diff(self.actions)
        assert diff.stack_name == self.external_name

    def test_diff__returns_generated_config(self):
        diff = self.differ.diff(self.actions)
        assert diff.generated_config == self.expected_generated_config

    def test_diff__returns_generated_template(self):
        diff = self.differ.diff(self.actions)
        assert diff.generated_template == self.actions.generate.return_value

    def test_diff__deployed_stack_exists__returns_is_deployed_as_true(self):
        diff = self.differ.diff(self.actions)
        assert diff.is_deployed is True

    def test_diff__deployed_stack_does_not_exist__returns_is_deployed_as_false(self):
        self.actions.describe.return_value = self.actions.describe.side_effect = None
        diff = self.differ.diff(self.actions)
        assert diff.is_deployed is False

    def test_diff__deployed_stack_does_not_exist__compares_none_to_generated_config(self):
        self.actions.describe.return_value = self.actions.describe.side_effect = None
        self.differ.diff(self.actions)

        self.command_capturer.compare_stack_configurations.assert_called_with(
            None,
            self.expected_generated_config
        )

    def test_diff__deployed_stack_does_not_exist__compares_empty_dict_string_to_generated_template(self):
        self.actions.fetch_remote_template.return_value = None
        self.differ.diff(self.actions)

        self.command_capturer.compare_templates.assert_called_with(
            '{}',
            self.actions.generate.return_value
        )

    @pytest.mark.parametrize(
        'status',
        [
            pytest.param(status)
            for status in [
                'CREATE_FAILED',
                'ROLLBACK_COMPLETE',
                'DELETE_COMPLETE',
            ]
        ]
    )
    def test_diff__non_deployed_stack_status__compares_none_to_generated_config(self, status):
        self.stack_status = status
        self.differ.diff(self.actions)

        self.command_capturer.compare_stack_configurations.assert_called_with(
            None,
            self.expected_generated_config
        )

    @pytest.mark.parametrize(
        'status',
        [
            pytest.param(status)
            for status in [
                'CREATE_FAILED',
                'ROLLBACK_COMPLETE',
                'DELETE_COMPLETE',
            ]
        ]
    )
    def test_diff__non_deployed_stack_status__compares_empty_dict_string_to_generated_template(self, status):
        self.stack_status = status
        self.differ.diff(self.actions)
        self.command_capturer.compare_templates.assert_called_with(
            '{}',
            self.actions.generate.return_value
        )

    def test_diff__deployed_stack_has_default_values__doesnt_pass_parameter__compares_identical_configs(self):
        self.deployed_parameters['new'] = 'default value'
        self.deployed_parameter_defaults['new'] = 'default value'
        self.differ.diff(self.actions)
        self.command_capturer.compare_stack_configurations.assert_called_with(
            self.expected_generated_config,
            self.expected_generated_config
        )

    def test_diff__deployed_stack_has_list_default_parameter__doesnt_pass_parameter__compares_identical_configs(self):
        self.deployed_parameters['new'] = 'first,second,third'
        self.deployed_parameter_defaults['new'] = 'first, second, third'
        self.deployed_parameter_types['new'] = 'CommaDelimitedList'
        self.differ.diff(self.actions)

        self.command_capturer.compare_stack_configurations.assert_called_with(
            self.expected_generated_config,
            self.expected_generated_config
        )

    def test_diff__deployed_stack_has_default_values__passes_the_parameter__compares_identical_configs(self):
        self.deployed_parameters['new'] = 'default value'
        self.deployed_parameter_defaults['new'] = 'default value'
        self.parameters_on_stack_config['new'] = 'default value'
        self.differ.diff(self.actions)
        self.command_capturer.compare_stack_configurations.assert_called_with(
            self.expected_generated_config,
            self.expected_generated_config
        )

    def test_diff__deployed_stack_has_default_values__passes_different_value__compares_different_configs(self):
        self.deployed_parameters['new'] = 'default value'
        self.deployed_parameter_defaults['new'] = 'default value'
        self.parameters_on_stack_config['new'] = 'custom value'
        self.differ.diff(self.actions)
        self.command_capturer.compare_stack_configurations.assert_called_with(
            self.expected_deployed_config,
            self.expected_generated_config
        )

    def test_diff__stack_exists_with_same_config_but_template_does_not__compares_identical_configs(self):
        self.actions.fetch_remote_template_summary.side_effect = None
        self.actions.fetch_remote_template_summary.return_value = None
        self.actions.fetch_remote_template.return_value = None
        self.differ.diff(self.actions)
        self.command_capturer.compare_stack_configurations.assert_called_with(
            self.expected_generated_config,
            self.expected_generated_config
        )

    def test_diff__deployed_parameter_has_linebreak_but_otherwise_no_difference__compares_identical_configs(self):
        self.deployed_parameters['param'] = self.deployed_parameters['param'] + '\n'
        self.differ.diff(self.actions)
        self.command_capturer.compare_stack_configurations.assert_called_with(
            self.expected_generated_config,
            self.expected_generated_config
        )

    def test_diff__parameter_has_identical_string_linebreak__compares_identical_configs(self):
        self.deployed_parameters['param'] = self.deployed_parameters['param'] + '\n'
        self.parameters_on_stack_config['param'] = self.parameters_on_stack_config['param'] + '\n'

        self.differ.diff(self.actions)
        generated_config = deepcopy(self.expected_generated_config._asdict())
        generated_parameters = generated_config.pop('parameters')
        expected_config = StackConfiguration(
            parameters={
                key: value.rstrip('\n')
                for key, value in generated_parameters.items()
            },
            **generated_config,
        )

        self.command_capturer.compare_stack_configurations.assert_called_with(
            expected_config,
            expected_config
        )

    def test_diff__parameter_has_identical_list_linebreaks__compares_identical_configs(self):
        self.deployed_parameter_types['param'] = 'CommaDelimitedList'
        self.deployed_parameters['param'] = 'testing\n,this\n,out\n'
        self.parameters_on_stack_config['param'] = [
            'testing\n',
            'this\n',
            'out\n'
        ]

        self.differ.diff(self.actions)
        generated_config = deepcopy(self.expected_generated_config._asdict())
        generated_parameters = generated_config.pop('parameters')
        generated_parameters['param'] = 'testing,this,out'
        expected_config = StackConfiguration(
            parameters=generated_parameters,
            **generated_config,
        )

        self.command_capturer.compare_stack_configurations.assert_called_with(
            expected_config,
            expected_config
        )

    def test_diff__no_echo_default_parameter__generated_stack_doesnt_pass_parameter__compares_identical_configs(self):
        self.deployed_parameters['new'] = '****'
        self.deployed_parameter_defaults['new'] = 'default value'
        self.deployed_no_echo_parameters.append('new')
        self.differ.diff(self.actions)
        self.command_capturer.compare_stack_configurations.assert_called_with(
            self.expected_generated_config,
            self.expected_generated_config
        )

    def test_diff__generated_template_has_no_echo_parameter__masks_value(self):
        self.parameters_on_stack_config['hide_me'] = "don't look at me!"
        self.local_no_echo_parameters.append('hide_me')

        expected_generated_config = self.expected_generated_config
        expected_generated_config.parameters['hide_me'] = StackDiffer.NO_ECHO_REPLACEMENT

        self.differ.diff(self.actions)

        self.command_capturer.compare_stack_configurations.assert_called_with(
            self.expected_deployed_config,
            expected_generated_config,
        )

    def test_diff__generated_template_has_no_echo_parameter__show_no_echo__shows_value(self):
        self.parameters_on_stack_config['hide_me'] = "don't look at me!"
        self.local_no_echo_parameters.append('hide_me')
        self.differ.show_no_echo = True
        self.differ.diff(self.actions)

        self.command_capturer.compare_stack_configurations.assert_called_with(
            self.expected_deployed_config,
            self.expected_generated_config,
        )


class TestDeepDiffStackDiffer:

    def setup_method(self, method):
        self.differ = DeepDiffStackDiffer()

        self.config1 = StackConfiguration(
            stack_name='stack',
            parameters={'pk1': 'pv1'},
            stack_tags={'tk1': 'tv1'},
            notifications=['notification'],
            role_arn=None
        )

        self.config2 = StackConfiguration(
            stack_name='stack',
            parameters={'pk1': 'pv1', 'pk2': 'pv2'},
            stack_tags={'tk1': 'tv1'},
            notifications=['notification'],
            role_arn='new_role'
        )

        self.template_dict_1 = {
            'AWSTemplateFormat': '2010-09-09',
            'Description': 'deployed',
            'Parameters': {'pk1': 'pv1'},
            'Resources': {}
        }
        self.template_dict_2 = {
            'AWSTemplateFormat': '2010-09-09',
            'Description': 'deployed',
            'Parameters': {'pk1': 'pv1'},
            'Resources': {
                'MyBucket': {
                    'Type': 'AWS::S3::Bucket',
                    'Properties': {
                        'BucketName': 'test'
                    }
                }
            }
        }

    def test_compare_stack_configurations__returns_deepdiff_of_deployed_and_generated(self):
        comparison = self.differ.compare_stack_configurations(self.config1, self.config2)
        assert comparison.t1 == self.config1
        assert comparison.t2 == self.config2

    def test_compare_stack_configurations__returned_deepdiff_has_verbosity_of_2(self):
        comparison = self.differ.compare_stack_configurations(self.config1, self.config2)
        assert comparison.verbose_level == 2

    def test_compare_stack_configurations__deployed_is_none__returns_deepdiff_with_none_for_t1(self):
        comparison = self.differ.compare_stack_configurations(None, self.config2)
        assert comparison.t1 is None

    @pytest.mark.parametrize(
        't1_serializer, t2_serializer',
        [
            pytest.param(json.dumps, json.dumps, id='templates are json'),
            pytest.param(yaml.dump, yaml.dump, id='templates are yaml'),
            pytest.param(json.dumps, yaml.dump, id="templates are mixed formats"),
        ]
    )
    def test_compare_templates__templates_are_json__returns_deepdiff_of_dicts(
        self,
        t1_serializer,
        t2_serializer
    ):
        template1, template2 = t1_serializer(self.template_dict_1), t2_serializer(self.template_dict_2)
        comparison = self.differ.compare_templates(template1, template2)
        assert comparison.t1 == self.template_dict_1
        assert comparison.t2 == self.template_dict_2

    def test_compare_templates__templates_are_yaml_with_intrinsic_functions__returns_deepdiff_of_dicts(self):
        template = """
            Resources:
              MyBucket:
                Type: AWS::S3::Bucket
                Properties:
                  BucketName: !Ref MyParam
        """
        comparison = self.differ.compare_templates(template, template)

        expected = cfn_flip.load_yaml(template)
        assert (comparison.t1, comparison.t2) == (expected, expected)

    def test_compare_templates__deployed_is_empty_dict_string__returns_deepdiff_with_empty_dict_for_t1(self):
        template = json.dumps(self.template_dict_1)
        comparison = self.differ.compare_templates('{}', template)
        assert comparison.t1 == {}


class TestDifflibStackDiffer:

    def setup_method(self, method):
        self.serialize = cfn_flip.dump_yaml
        self.differ = DifflibStackDiffer()
        self.config1 = StackConfiguration(
            stack_name='stack',
            parameters={'pk1': 'pv1'},
            stack_tags={'tk1': 'tv1'},
            notifications=['notification'],
            role_arn=None
        )

        self.config2 = StackConfiguration(
            stack_name='stack',
            parameters={'pk1': 'pv1', 'pk2': 'pv2'},
            stack_tags={'tk1': 'tv1'},
            notifications=['notification'],
            role_arn='new_role'
        )

        self.template_dict_1 = {
            'AWSTemplateFormat': '2010-09-09',
            'Description': 'deployed',
            'Parameters': {'pk1': 'pv1'},
            'Resources': {}
        }
        self.template_dict_2 = {
            'AWSTemplateFormat': '2010-09-09',
            'Description': 'deployed',
            'Parameters': {'pk1': 'pv1'},
            'Resources': {
                'MyBucket': {
                    'Type': 'AWS::S3::Bucket',
                    'Properties': {
                        'BucketName': 'test'
                    }
                }
            }
        }

    def create_expected_diff(self, first, second, already_formatted=False):
        if not already_formatted:
            deployed_dict, deployed_format = cfn_flip.load(first)
            generated_dict, generated_format = cfn_flip.load(second)
            dumpers = {
                'json': cfn_flip.dump_json,
                'yaml': cfn_flip.dump_yaml
            }
            first = dumpers[generated_format](deployed_dict)
            second = dumpers[generated_format](generated_dict)
        first_list, second_list = first.splitlines(), second.splitlines()
        return list(difflib.unified_diff(
            first_list,
            second_list,
            fromfile='deployed',
            tofile='generated',
            lineterm=''
        ))

    def test_compare_stack_configurations__returns_diff_of_deployed_and_generated_when_converted_to_dicts(self):
        comparison = self.differ.compare_stack_configurations(self.config1, self.config2)
        expected_config_1_dict = self.make_config_comparable(self.config1)
        expected_config_2_dict = self.make_config_comparable(self.config2)
        expected_config_1 = self.serialize(expected_config_1_dict)
        expected_config_2 = self.serialize(expected_config_2_dict)
        expected = self.create_expected_diff(expected_config_1, expected_config_2)

        assert comparison == expected

    def make_config_comparable(self, config: StackConfiguration):
        config_dict = dict(config._asdict())
        without_empty_values = {
            key: value
            for key, value in config_dict.items()
            if value not in (None, [], {}) and key != 'stack_name'
        }
        return without_empty_values

    def test_compare_stack_configurations__deployed_is_none__returns_diff_with_none(self):
        comparison = self.differ.compare_stack_configurations(None, self.config2)
        expected = self.create_expected_diff(
            self.serialize(None),
            self.serialize(self.make_config_comparable(self.config2))
        )
        assert comparison == expected

    def test_compare_stack_configurations__deployed_is_none__all_configs_are_falsey__returns_diff_with_none(self):
        empty_config = StackConfiguration(
            stack_name='stack',
            parameters={},
            stack_tags={},
            notifications=[],
            role_arn=None
        )
        comparison = self.differ.compare_stack_configurations(None, empty_config)

        expected = self.create_expected_diff(
            self.serialize(None),
            self.serialize(self.make_config_comparable(empty_config)),
            already_formatted=True
        )
        assert comparison == expected

    @pytest.mark.parametrize(
        'serializer',
        [
            pytest.param(json.dumps, id='templates are json'),
            pytest.param(yaml.dump, id='templates are yaml'),
        ]
    )
    def test_compare_templates__templates_are_json__returns_deepdiff_of_dicts(
        self,
        serializer,
    ):
        template1, template2 = serializer(self.template_dict_1), serializer(self.template_dict_2)
        comparison = self.differ.compare_templates(template1, template2)
        expected = self.create_expected_diff(template1, template2)
        assert comparison == expected

    def test_compare_templates__deployed_is_empty_dict_string__returns_diff_with_empty_string(self):
        template = json.dumps(self.template_dict_1)
        comparison = self.differ.compare_templates('{}', template)
        expected = self.create_expected_diff('{}', template)
        assert comparison == expected

    def test_compare_templates__json_template__only_indentation_diff__returns_no_diff(self):
        template1 = json.dumps(self.template_dict_1, indent=2)
        template2 = json.dumps(self.template_dict_1, indent=4)
        comparison = self.differ.compare_templates(template1, template2)
        assert len(comparison) == 0

    def test_compare_templates__yaml_template__only_indentation_diff__returns_no_diff(self):
        template1 = yaml.dump(self.template_dict_1, indent=2)
        template2 = yaml.dump(self.template_dict_1, indent=4)
        comparison = self.differ.compare_templates(template1, template2)
        assert len(comparison) == 0

    def test_compare_templates__opposite_template_types_but_identical_template__returns_no_diff(self):
        template1 = json.dumps(self.template_dict_1)
        template2 = yaml.dump(self.template_dict_1)
        comparison = self.differ.compare_templates(template1, template2)
        assert len(comparison) == 0



# -*- coding: utf-8 -*-

import logging
import warnings


__author__ = 'Cloudreach'
__email__ = 'sceptre@cloudreach.com'
__version__ = '3.2.2'


# Set up logging to ``/dev/null`` like a library is supposed to.
# http://docs.python.org/3.3/howto/logging.html#configuring-logging-for-a-library
class NullHandler(logging.Handler):  # pragma: no cover
    def emit(self, record):
        pass


with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)

logging.getLogger('sceptre').addHandler(NullHandler())



# -*- coding: utf-8 -*-

import os
import pytest
from unittest.mock import Mock, MagicMock, patch, sentinel, ANY
from moto import mock_s3

from boto3.session import Session
from botocore.exceptions import ClientError, UnknownServiceError

from sceptre.connection_manager import ConnectionManager, _retry_boto_call
from sceptre.exceptions import RetryLimitExceededError, InvalidAWSCredentialsError


class TestConnectionManager(object):

    def setup_method(self, test_method):
        self.stack_name = None
        self.profile = None
        self.iam_role = None
        self.iam_role_session_duration = 3600
        self.region = "eu-west-1"

        ConnectionManager._boto_sessions = {}
        ConnectionManager._clients = {}
        ConnectionManager._stack_keys = {}

        # Temporary workaround for https://github.com/spulec/moto/issues/1924
        os.environ.setdefault("AWS_ACCESS_KEY_ID", "sceptre_test_key_id")
        os.environ.setdefault("AWS_SECRET_ACCESS_KEY", "sceptre_test_access_key")

        self.connection_manager = ConnectionManager(
            region=self.region,
            stack_name=self.stack_name,
            profile=self.profile,
            iam_role=self.iam_role
        )

    def test_connection_manager_initialised_with_no_optional_parameters(self):
        connection_manager = ConnectionManager(region=sentinel.region)

        assert connection_manager.stack_name is None
        assert connection_manager.profile is None
        assert connection_manager.region == sentinel.region
        assert connection_manager._boto_sessions == {}
        assert connection_manager._clients == {}
        assert connection_manager._stack_keys == {}

    def test_connection_manager_initialised_with_all_parameters(self):
        connection_manager = ConnectionManager(
            region=self.region,
            stack_name="stack",
            profile="profile",
            iam_role="iam_role",
            iam_role_session_duration=21600
        )

        assert connection_manager.stack_name == "stack"
        assert connection_manager.profile == "profile"
        assert connection_manager.iam_role == "iam_role"
        assert connection_manager.iam_role_session_duration == 21600
        assert connection_manager.region == self.region
        assert connection_manager._boto_sessions == {}
        assert connection_manager._clients == {}
        assert connection_manager._stack_keys == {
            "stack": (self.region, "profile", "iam_role")
        }

    def test_repr(self):
        self.connection_manager.stack_name = "stack"
        self.connection_manager.profile = "profile"
        self.connection_manager.region = "region"
        self.connection_manager.iam_role = "iam_role"
        response = self.connection_manager.__repr__()
        assert response == "sceptre.connection_manager.ConnectionManager(" \
            "region='region', profile='profile', stack_name='stack', "\
            "iam_role='iam_role', iam_role_session_duration='None')"

    def test_repr_with_iam_role_session_duration(self):
        self.connection_manager.stack_name = "stack"
        self.connection_manager.profile = "profile"
        self.connection_manager.region = "region"
        self.connection_manager.iam_role = "iam_role"
        self.connection_manager.iam_role_session_duration = 21600
        response = self.connection_manager.__repr__()
        assert response == "sceptre.connection_manager.ConnectionManager(" \
            "region='region', profile='profile', stack_name='stack', "\
            "iam_role='iam_role', iam_role_session_duration='21600')"

    def test_boto_session_with_cache(self):
        self.connection_manager._boto_sessions["test"] = sentinel.boto_session

        boto_session = self.connection_manager._boto_sessions["test"]
        assert boto_session == sentinel.boto_session

    @patch("sceptre.connection_manager.boto3.session.Session")
    def test_boto_session_with_no_profile(
            self, mock_Session
    ):
        self.connection_manager._boto_sessions = {}
        self.connection_manager.profile = None

        boto_session = self.connection_manager._get_session(
            self.connection_manager.profile, self.region, self.iam_role
        )

        assert boto_session.isinstance(mock_Session)
        mock_Session.assert_called_once_with(
            profile_name=None,
            region_name="eu-west-1",
            aws_access_key_id=ANY,
            aws_secret_access_key=ANY,
            aws_session_token=ANY
        )

    @patch("sceptre.connection_manager.boto3.session.Session")
    def test_boto_session_with_profile(self, mock_Session):
        self.connection_manager._boto_sessions = {}
        self.connection_manager.profile = "profile"

        boto_session = self.connection_manager._get_session(
            self.connection_manager.profile, self.region, self.iam_role
        )

        assert boto_session.isinstance(mock_Session)
        mock_Session.assert_called_once_with(
            profile_name="profile",
            region_name="eu-west-1",
            aws_access_key_id=ANY,
            aws_secret_access_key=ANY,
            aws_session_token=ANY
        )

    @patch("sceptre.connection_manager.boto3.session.Session")
    def test_boto_session_with_no_iam_role(
            self, mock_Session
    ):
        self.connection_manager._boto_sessions = {}
        self.connection_manager.iam_role = None

        boto_session = self.connection_manager._get_session(
            self.profile, self.region, self.connection_manager.iam_role
        )

        assert boto_session.isinstance(mock_Session)
        mock_Session.assert_called_once_with(
            profile_name=None,
            region_name="eu-west-1",
            aws_access_key_id=ANY,
            aws_secret_access_key=ANY,
            aws_session_token=ANY
        )

        boto_session.client().assume_role.assert_not_called()

    @patch("sceptre.connection_manager.boto3.session.Session")
    def test_boto_session_with_iam_role(self, mock_Session):
        self.connection_manager._boto_sessions = {}
        self.connection_manager.iam_role = "iam_role"

        boto_session = self.connection_manager._get_session(
            self.profile, self.region, self.connection_manager.iam_role
        )

        assert boto_session.isinstance(mock_Session)
        mock_Session.assert_any_call(
            profile_name=None,
            region_name="eu-west-1",
            aws_access_key_id=ANY,
            aws_secret_access_key=ANY,
            aws_session_token=ANY
        )

        boto_session.client().assume_role.assert_called_once_with(
            RoleArn=self.connection_manager.iam_role,
            RoleSessionName="{0}-session".format(
                self.connection_manager.iam_role.split("/")[-1]
            )
        )

        credentials = boto_session.client().assume_role()["Credentials"]

        mock_Session.assert_any_call(
            region_name="eu-west-1",
            aws_access_key_id=credentials["AccessKeyId"],
            aws_secret_access_key=credentials["SecretAccessKey"],
            aws_session_token=credentials["SessionToken"]
        )

    @patch("sceptre.connection_manager.boto3.session.Session")
    def test_boto_session_with_iam_role_session_duration(self, mock_Session):
        self.connection_manager._boto_sessions = {}
        self.connection_manager.iam_role = "iam_role"
        self.connection_manager.iam_role_session_duration = 21600

        boto_session = self.connection_manager._get_session(
            self.profile, self.region, self.connection_manager.iam_role
        )

        boto_session.client().assume_role.assert_called_once_with(
            RoleArn=self.connection_manager.iam_role,
            RoleSessionName="{0}-session".format(
                self.connection_manager.iam_role.split("/")[-1]
            ),
            DurationSeconds=21600
        )

    @patch("sceptre.connection_manager.boto3.session.Session")
    def test_boto_session_with_iam_role_returning_empty_credentials(self, mock_Session):
        self.connection_manager._boto_sessions = {}
        self.connection_manager.iam_role = "iam_role"

        mock_Session.return_value.get_credentials.side_effect = [
            MagicMock(), None, MagicMock(), MagicMock(), MagicMock()
        ]

        with pytest.raises(InvalidAWSCredentialsError):
            self.connection_manager._get_session(
                self.profile, self.region, self.connection_manager.iam_role
            )

    @patch("sceptre.connection_manager.boto3.session.Session")
    def test_two_boto_sessions(self, mock_Session):
        self.connection_manager._boto_sessions = {
            "one": mock_Session,
            "two": mock_Session
        }

        boto_session_1 = self.connection_manager._boto_sessions["one"]
        boto_session_2 = self.connection_manager._boto_sessions["two"]
        assert boto_session_1 == boto_session_2

    @patch("sceptre.connection_manager.boto3.session.Session.get_credentials")
    def test_get_client_with_no_pre_existing_clients(
        self, mock_get_credentials
    ):
        service = "s3"
        region = "eu-west-1"
        profile = None
        iam_role = None
        stack = self.stack_name

        client = self.connection_manager._get_client(
            service, region, profile, stack, iam_role
        )
        expected_client = Session().client(service)
        assert str(type(client)) == str(type(expected_client))

    @patch("sceptre.connection_manager.boto3.session.Session.get_credentials")
    def test_get_client_with_invalid_client_type(self, mock_get_credentials):
        service = "invalid_type"
        region = "eu-west-1"
        iam_role = None
        profile = None
        stack = self.stack_name

        with pytest.raises(UnknownServiceError):
            self.connection_manager._get_client(
                service, region, profile, stack, iam_role
            )

    @patch("sceptre.connection_manager.boto3.session.Session.get_credentials")
    def test_get_client_with_exisiting_client(self, mock_get_credentials):
        service = "cloudformation"
        region = "eu-west-1"
        iam_role = None
        profile = None
        stack = self.stack_name

        client_1 = self.connection_manager._get_client(
            service, region, profile, stack, iam_role
        )
        client_2 = self.connection_manager._get_client(
            service, region, profile, stack, iam_role
        )
        assert client_1 == client_2

    @patch("sceptre.connection_manager.boto3.session.Session.get_credentials")
    def test_get_client_with_exisiting_client_and_profile_none(
            self, mock_get_credentials
    ):
        service = "cloudformation"
        region = "eu-west-1"
        iam_role = None
        profile = None
        stack = self.stack_name

        self.connection_manager.profile = None
        client_1 = self.connection_manager._get_client(
            service, region, profile, stack, iam_role
        )
        client_2 = self.connection_manager._get_client(
            service, region, profile, stack, iam_role
        )
        assert client_1 == client_2

    @mock_s3
    def test_call_with_valid_service_and_call(self):
        service = 's3'
        command = 'list_buckets'

        return_value = self.connection_manager.call(service, command, {})
        assert return_value['ResponseMetadata']['HTTPStatusCode'] == 200

    @mock_s3
    def test_call_with_valid_service_and_stack_name_call(self):
        service = 's3'
        command = 'list_buckets'

        connection_manager = ConnectionManager(
            region=self.region,
            stack_name='stack'
        )

        return_value = connection_manager.call(
            service, command, {}, stack_name='stack'
        )
        assert return_value['ResponseMetadata']['HTTPStatusCode'] == 200


class TestRetry():

    def test_retry_boto_call_returns_response_correctly(self):
        def func(*args, **kwargs):
            return sentinel.response

        response = _retry_boto_call(func)()

        assert response == sentinel.response

    @patch("sceptre.connection_manager.time.sleep")
    def test_retry_boto_call_pauses_when_request_limit_hit(
            self, mock_sleep
    ):
        mock_func = Mock()
        mock_func.side_effect = [
            ClientError(
                {
                    "Error": {
                        "Code": "Throttling",
                        "Message": "Request limit hit"
                    }
                },
                sentinel.operation
            ),
            sentinel.response
        ]
        # The attribute function.__name__ is required by the decorator @wraps.
        mock_func.__name__ = "mock_func"

        _retry_boto_call(mock_func)()
        mock_sleep.assert_called_once_with(1)

    def test_retry_boto_call_raises_non_throttling_error(self):
        mock_func = Mock()
        mock_func.side_effect = ClientError(
            {
                "Error": {
                    "Code": 500,
                    "Message": "Boom!"
                }
            },
            sentinel.operation
        )
        # The attribute function.__name__ is required by the decorator @wraps.
        mock_func.__name__ = "mock_func"

        with pytest.raises(ClientError) as e:
            _retry_boto_call(mock_func)()
        assert e.value.response["Error"]["Code"] == 500
        assert e.value.response["Error"]["Message"] == "Boom!"

    @patch("sceptre.connection_manager.time.sleep")
    def test_retry_boto_call_raises_retry_limit_exceeded_exception(
            self, mock_sleep
    ):
        mock_func = Mock()
        mock_func.side_effect = ClientError(
            {
                "Error": {
                    "Code": "Throttling",
                    "Message": "Request limit hit"
                }
            },
            sentinel.operation
        )
        # The attribute function.__name__ is required by the decorator @wraps.
        mock_func.__name__ = "mock_func"

        with pytest.raises(RetryLimitExceededError):
            _retry_boto_call(mock_func)()



import functools
import itertools
from collections import defaultdict
from typing import Set, Optional, List
from unittest.mock import Mock, create_autospec

import pytest

from sceptre.cli.prune import Pruner, PATH_FOR_WHOLE_PROJECT
from sceptre.context import SceptreContext
from sceptre.exceptions import CannotPruneStackError
from sceptre.plan.plan import SceptrePlan
from sceptre.stack import Stack
from sceptre.stack_status import StackStatus


class FakePlan(SceptrePlan):
    def __init__(
        self,
        context: SceptreContext,
        command_stacks: Set[Stack],
        all_stacks: Set[Stack],
        statuses_to_return: dict,
    ):
        self.context = context
        self.command = None
        self.reverse = None
        self.launch_order: Optional[List[Set[Stack]]] = None

        self.all_stacks = self.graph = all_stacks
        self.command_stacks = command_stacks
        self.statuses_to_return = statuses_to_return

        self.executions = []

    def _execute(self, *args):
        self.executions.append(
            (self.command, self.launch_order.copy(), args)
        )
        return {
            stack: self.statuses_to_return[stack]
            for stack in self
        }

    def _generate_launch_order(self, reverse=False) -> List[Set[Stack]]:
        launch_order = [self.command_stacks]
        if self.context.ignore_dependencies:
            return launch_order

        all_stacks = list(self.all_stacks)

        for start_index in range(0, len(all_stacks), 2):
            chunk = {
                stack for stack in all_stacks[start_index: start_index + 2]
                if stack not in self.command_stacks
                and self._has_dependency_on_a_command_stack(stack)
            }
            if len(chunk):
                launch_order.append(chunk)

        return launch_order

    @functools.lru_cache()
    def _has_dependency_on_a_command_stack(self, stack):
        if len(self.command_stacks.intersection(stack.dependencies)):
            return True

        for dependency in stack.dependencies:
            if self._has_dependency_on_a_command_stack(dependency):
                return True

        return False


class TestPruner:

    def setup_method(self, test_method):
        self.plans: List[FakePlan] = []

        self.context = SceptreContext(
            project_path="project",
            command_path=PATH_FOR_WHOLE_PROJECT,
        )

        self.all_stacks = [
            Mock(spec=Stack, ignore=False, obsolete=False, dependencies=[]),
            Mock(spec=Stack, ignore=False, obsolete=False, dependencies=[]),
            Mock(spec=Stack, ignore=False, obsolete=False, dependencies=[]),
            Mock(spec=Stack, ignore=False, obsolete=False, dependencies=[]),
            Mock(spec=Stack, ignore=False, obsolete=False, dependencies=[]),
            Mock(spec=Stack, ignore=False, obsolete=False, dependencies=[])
        ]
        for index, stack in enumerate(self.all_stacks):
            stack.name = f'stacks/stack-{index}.yaml'

        self.command_stacks = [
            self.all_stacks[2],
            self.all_stacks[4]
        ]

        self.statuses_to_return = defaultdict(lambda: StackStatus.COMPLETE)

        self.plan_factory = create_autospec(SceptrePlan)
        self.plan_factory.side_effect = self.fake_plan_factory

        self.pruner = Pruner(self.context, self.plan_factory)

    def fake_plan_factory(self, sceptre_context):
        fake_plan = FakePlan(
            sceptre_context,
            set(self.command_stacks),
            set(self.all_stacks),
            self.statuses_to_return
        )
        self.plans.append(fake_plan)
        return fake_plan

    @property
    def executed_stacks(self):
        assert len(self.plans) == 1
        launch_order = self.plans[0].executions[0][1]
        return list(itertools.chain.from_iterable(launch_order))

    def test_prune__no_obsolete_stacks__returns_zero(self):
        code = self.pruner.prune()
        assert code == 0

    def test_prune__no_obsolete_stacks__does_not_call_command_on_plan(self):
        self.pruner.prune()
        assert len(self.plans[0].executions) == 0

    def test_prune__whole_project__obsolete_stacks__deletes_all_obsolete_stacks(self):
        self.all_stacks[4].obsolete = True
        self.all_stacks[5].obsolete = True

        self.pruner.prune()

        assert self.plans[0].executions[0][0] == 'delete'
        assert set(self.executed_stacks) == {self.all_stacks[4], self.all_stacks[5]}

    def test_prune__command_path__obsolete_stacks__deletes_only_obsolete_stacks_on_path(self):
        self.all_stacks[4].obsolete = True  # On command path
        self.all_stacks[5].obsolete = True  # not on command path
        self.context.command_path = "my/command/path"
        self.pruner.prune()

        assert self.plans[0].executions[0][0] == 'delete'
        assert set(self.executed_stacks) == {self.all_stacks[4]}

    def test_prune__obsolete_stacks__returns_zero(self):
        self.all_stacks[4].obsolete = True
        self.all_stacks[5].obsolete = True

        code = self.pruner.prune()
        assert code == 0

    def test_prune__obsolete_stacks_depend_on_other_obsolete_stacks__deletes_only_obsolete_stacks(self):
        self.all_stacks[1].obsolete = True
        self.all_stacks[3].obsolete = True
        self.all_stacks[4].obsolete = True
        self.all_stacks[5].obsolete = True
        self.all_stacks[3].dependencies.append(self.all_stacks[1])
        self.all_stacks[4].dependencies.append(self.all_stacks[3])
        self.all_stacks[5].dependencies.append(self.all_stacks[3])

        self.pruner.prune()

        assert self.plans[0].executions[0][0] == 'delete'
        assert set(self.executed_stacks) == {
            self.all_stacks[1],
            self.all_stacks[3],
            self.all_stacks[4],
            self.all_stacks[5],
        }

    def test_prune__non_obsolete_stacks_depend_on_obsolete_stacks__raises_cannot_prune_stack_error(self):
        self.all_stacks[1].obsolete = True
        self.all_stacks[3].obsolete = False
        self.all_stacks[4].obsolete = False
        self.all_stacks[5].obsolete = False
        self.all_stacks[3].dependencies.append(self.all_stacks[1])
        self.all_stacks[4].dependencies.append(self.all_stacks[3])
        self.all_stacks[5].dependencies.append(self.all_stacks[3])

        with pytest.raises(CannotPruneStackError):
            self.pruner.prune()

    def test_prune__non_obsolete_stacks_depend_on_obsolete_stacks__ignore_dependencies__deletes_obsolete_stacks(self):
        self.all_stacks[1].obsolete = True
        self.all_stacks[3].obsolete = False
        self.all_stacks[4].obsolete = False
        self.all_stacks[5].obsolete = False
        self.all_stacks[3].dependencies.append(self.all_stacks[1])
        self.all_stacks[4].dependencies.append(self.all_stacks[3])
        self.all_stacks[5].dependencies.append(self.all_stacks[3])
        self.context.ignore_dependencies = True
        self.pruner.prune()

        assert self.plans[0].executions[0][0] == 'delete'
        assert set(self.executed_stacks) == {
            self.all_stacks[1],
        }

    def test_prune__delete_action_fails__returns_nonzero(self):
        self.all_stacks[1].obsolete = True
        self.statuses_to_return[self.all_stacks[1]] = StackStatus.FAILED

        code = self.pruner.prune()
        assert code != 0



import difflib
from copy import deepcopy
from io import StringIO
from typing import TextIO
from unittest.mock import Mock

import cfn_flip
import pytest
import yaml
from deepdiff import DeepDiff

from sceptre.diffing.diff_writer import DiffWriter, DeepDiffWriter, deepdiff_json_defaults, DiffLibWriter
from sceptre.diffing.stack_differ import StackDiff, DiffType, StackConfiguration


class ImplementedDiffWriter(DiffWriter):

    def __init__(
        self,
        stack_diff: StackDiff,
        output_stream: TextIO,
        output_format: str,
        capturing_mock: Mock
    ):
        super().__init__(stack_diff, output_stream, output_format)
        self.capturing_mock = capturing_mock

    def dump_diff(self, diff: DiffType) -> str:
        return self.capturing_mock.dump_diff(diff)

    @property
    def has_config_difference(self) -> bool:
        return self.capturing_mock.has_config_difference

    @property
    def has_template_difference(self) -> bool:
        return self.capturing_mock.has_template_difference


class TestDiffWriter:

    def setup_method(self, method):
        self.diff_output = 'diff'
        self.capturing_mock = Mock(**{
            'dump_diff.return_value': self.diff_output
        })
        self.stack_name = 'stack'
        self.template_diff = Mock()
        self.config_diff = Mock()
        self.is_deployed = True
        self.generated_template = 'my template'
        self.output_format = 'yaml'

        self.output_stream = StringIO()

        self.diff_detected_message = f'--> Difference detected for stack {self.stack_name}!'

    @property
    def generated_config(self):
        return StackConfiguration(
            stack_name=self.stack_name,
            parameters={},
            stack_tags={},
            notifications=[],
            role_arn=None
        )

    @property
    def diff(self):
        return StackDiff(
            self.stack_name,
            self.template_diff,
            self.config_diff,
            self.is_deployed,
            self.generated_config,
            self.generated_template
        )

    @property
    def writer(self):
        return ImplementedDiffWriter(
            self.diff,
            self.output_stream,
            self.output_format,
            self.capturing_mock
        )

    def assert_expected_output(self, *expected_segments):
        expected_segments = [f'{line}\n' for line in expected_segments]
        joined = ''.join(expected_segments)
        expected_split_lines = joined.splitlines()

        received_lines = self.output_stream.getvalue().splitlines()
        diff = list(difflib.unified_diff(
            received_lines,
            expected_split_lines,
            fromfile='actual',
            tofile='expected'
        ))
        assert not diff, '\n'.join(diff)

    def test_write__no_difference__writes_no_difference(self):
        self.capturing_mock.has_config_difference = False
        self.capturing_mock.has_template_difference = False

        self.writer.write()

        self.assert_expected_output(
            DiffWriter.STAR_BAR,
            f'No difference to deployed stack {self.stack_name}'
        )

    @pytest.mark.parametrize(
        'output_format, config_serializer',
        [
            pytest.param('yaml', cfn_flip.dump_yaml, id='output format is yaml'),
            pytest.param('json', cfn_flip.dump_json, id='output format is json'),
            pytest.param('text', cfn_flip.dump_yaml, id='output format is text')
        ]
    )
    def test_write__new_stack__writes_new_stack_config_and_template(self, output_format, config_serializer):
        self.is_deployed = False
        self.output_format = output_format

        self.writer.write()

        self.assert_expected_output(
            DiffWriter.STAR_BAR,
            self.diff_detected_message,
            'This stack is not deployed yet!',
            DiffWriter.LINE_BAR,
            'New Config:',
            '',
            config_serializer(dict(self.generated_config._asdict())),
            DiffWriter.LINE_BAR,
            'New Template:',
            '',
            self.generated_template
        )

    def test_write__only_config_is_different__writes_config_difference(self):
        self.capturing_mock.has_config_difference = True
        self.capturing_mock.has_template_difference = False

        self.writer.write()

        self.assert_expected_output(
            DiffWriter.STAR_BAR,
            self.diff_detected_message,
            DiffWriter.LINE_BAR,
            f'Config difference for {self.stack_name}:',
            '',
            self.diff_output,
            DiffWriter.LINE_BAR,
            'No template difference'
        )

    def test_write__only_template_is_different__writes_template_difference(self):
        self.capturing_mock.has_config_difference = False
        self.capturing_mock.has_template_difference = True

        self.writer.write()

        self.assert_expected_output(
            DiffWriter.STAR_BAR,
            self.diff_detected_message,
            DiffWriter.LINE_BAR,
            'No stack config difference',
            DiffWriter.LINE_BAR,
            f'Template difference for {self.stack_name}:',
            '',
            self.diff_output
        )

    def test_write__config_and_template_are_different__writes_both_differences(self):
        self.capturing_mock.has_config_difference = True
        self.capturing_mock.has_template_difference = True

        self.writer.write()

        self.assert_expected_output(
            DiffWriter.STAR_BAR,
            self.diff_detected_message,
            DiffWriter.LINE_BAR,
            f'Config difference for {self.stack_name}:',
            '',
            self.diff_output,
            DiffWriter.LINE_BAR,
            f'Template difference for {self.stack_name}:',
            '',
            self.diff_output
        )


class TestDeepDiffWriter:
    def setup_method(self, method):
        self.stack_name = 'stack'

        self.is_deployed = True
        self.output_format = 'yaml'

        self.output_stream = StringIO()

        self.config1 = StackConfiguration(
            stack_name=self.stack_name,
            parameters={},
            stack_tags={},
            notifications=[],
            role_arn=None
        )

        self.config2 = deepcopy(self.config1)

        self.template1 = 'template'
        self.template2 = 'template'

    @property
    def template_diff(self):
        return DeepDiff(self.template1, self.template2)

    @property
    def config_diff(self):
        return DeepDiff(self.config1, self.config2)

    @property
    def diff(self):
        return StackDiff(
            self.stack_name,
            self.template_diff,
            self.config_diff,
            self.is_deployed,
            self.config1,
            self.template1
        )

    @property
    def writer(self):
        return DeepDiffWriter(
            self.diff,
            self.output_stream,
            self.output_format,
        )

    def test_has_config_difference__config_difference_is_present__returns_true(self):
        self.config2.parameters['new_key'] = 'new value'
        assert self.writer.has_config_difference

    def test_has_config_difference__config_difference_is_absent__returns_false(self):
        assert self.writer.has_config_difference is False

    def test_has_template_difference__template_difference_is_present__returns_true(self):
        self.template2 = 'new'
        assert self.writer.has_template_difference

    def test_has_template_difference__template_difference_is_absent__returns_false(self):
        assert self.writer.has_template_difference is False

    def test_dump_diff__output_format_is_json__outputs_to_json(self):
        self.output_format = 'json'
        self.config2.parameters['new_key'] = 'new value'

        result = self.writer.dump_diff(self.config_diff)
        expected = self.config_diff.to_json(indent=4, default_mapping=deepdiff_json_defaults)
        assert result == expected

    def test_dump_diff__output_format_is_yaml__outputs_to_yaml(self):
        self.output_format = 'yaml'
        self.config2.parameters['new_key'] = 'new value'

        result = self.writer.dump_diff(self.config_diff)
        expected_dict = self.config_diff.to_dict()
        expected_yaml = yaml.dump(expected_dict, indent=4)
        assert result == expected_yaml

    def test_dump_diff__output_format_is_text__outputs_to_yaml(self):
        self.output_format = 'text'
        self.config2.parameters['new_key'] = 'new value'

        result = self.writer.dump_diff(self.config_diff)
        expected_dict = self.config_diff.to_dict()
        expected_yaml = yaml.dump(expected_dict, indent=4)
        assert result == expected_yaml

    def test_dump_diff__output_format_is_yaml__diff_has_multiline_strings__strips_out_extra_spaces(self):
        self.config1.parameters['long_param'] = 'here \nis \nmy \nlong \nstring'
        self.config2.parameters['long_param'] = 'here \nis \nmy \nother \nlong \nstring'

        dumped = self.writer.dump_diff(self.config_diff)
        loaded = yaml.safe_load(dumped)
        assert ' ' not in loaded['values_changed']["root.parameters['long_param']"]['new_value']
        assert ' ' not in loaded['values_changed']["root.parameters['long_param']"]['old_value']
        expected_diff = '\n'.join(
            difflib.unified_diff(
                self.config1.parameters['long_param'].splitlines(),
                self.config2.parameters['long_param'].splitlines(),
                lineterm=''
            )
        ).replace(' \n', '\n')
        assert expected_diff == loaded['values_changed']["root.parameters['long_param']"]['diff']


class TestDiffLibWriter:
    def setup_method(self, method):
        self.stack_name = 'stack'

        self.is_deployed = True
        self.output_format = 'yaml'

        self.output_stream = StringIO()

        self.config1 = StackConfiguration(
            stack_name=self.stack_name,
            parameters={},
            stack_tags={},
            notifications=[],
            role_arn=None
        )

        self.config2 = deepcopy(self.config1)

        self.template1 = 'template'
        self.template2 = 'template'

    @property
    def template_diff(self):
        return list(difflib.unified_diff(self.template1, self.template2))

    @property
    def config_diff(self):
        config_1 = yaml.dump(dict(self.config1._asdict())).splitlines()
        config_2 = yaml.dump(dict(self.config2._asdict())).splitlines()
        return list(difflib.unified_diff(config_1, config_2))

    @property
    def diff(self):
        return StackDiff(
            self.stack_name,
            self.template_diff,
            self.config_diff,
            self.is_deployed,
            self.config1,
            self.template1
        )

    @property
    def writer(self):
        return DiffLibWriter(
            self.diff,
            self.output_stream,
            self.output_format,
        )

    def test_has_config_difference__config_difference_is_present__returns_true(self):
        self.config2.parameters['new_key'] = 'new value'
        assert self.writer.has_config_difference

    def test_has_config_difference__config_difference_is_absent__returns_false(self):
        assert self.writer.has_config_difference is False

    def test_has_template_difference__template_difference_is_present__returns_true(self):
        self.template2 = 'new'
        assert self.writer.has_template_difference

    def test_has_template_difference__template_difference_is_absent__returns_false(self):
        assert self.writer.has_template_difference is False

    def test_dump_diff__returns_joined_list(self):
        result = self.writer.dump_diff(self.diff.config_diff)
        expected = '\n'.join(self.diff.config_diff)
        assert result == expected



import logging
from typing import List

import click
from click import Context
from colorama import Fore, Style

from sceptre.cli.helpers import catch_exceptions, confirmation, stack_status_exit_code
from sceptre.cli.prune import Pruner
from sceptre.context import SceptreContext
from sceptre.exceptions import DependencyDoesNotExistError
from sceptre.plan.plan import SceptrePlan
from sceptre.stack import Stack

logger = logging.getLogger(__name__)


@click.command(name="launch", short_help="Launch a Stack or StackGroup.")
@click.argument("path")
@click.option("-y", "--yes", is_flag=True, help="Assume yes to all questions.")
@click.option(
    "-p",
    "--prune",
    is_flag=True,
    help="If set, will delete all stacks in the command path marked as obsolete."
)
@click.pass_context
@catch_exceptions
def launch_command(ctx: Context, path: str, yes: bool, prune: bool):
    """
    Launch a Stack or StackGroup for a given config PATH. This command is intended as a catch-all
    command that will apply any changes from Stack Configs indicated via the path.

    \b
    * Any Stacks that do not exist will be created
    * Any stacks that already exist will be updated (if there are any changes)
    * If any stacks are marked with "ignore: True", those stacks will neither be created nor updated
    * If any stacks are marked with "obsolete: True", those stacks will neither be created nor updated.
      Furthermore, if the "-p"/"--prune" flag is used, these stacks will be deleted prior to any
      other launch commands
    """
    context = SceptreContext(
        command_path=path,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        options=ctx.obj.get("options"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies")
    )
    launcher = Launcher(context)
    launcher.print_operations(prune)
    if not yes:
        launcher.confirm(prune)

    exit_code = launcher.launch(prune)
    exit(exit_code)


class Launcher:
    """Launcher is a utility to coordinate the flow of launching.

    :param context: The Sceptre context to use for launching
    :param plan_factory: A callable with the signature of (SceptreContext) -> SceptrePlan
    """
    def __init__(self, context: SceptreContext, plan_factory=SceptrePlan, pruner_factory=Pruner):
        self._context = context
        self._make_plan = plan_factory
        self._make_pruner = pruner_factory

        self._plan = None

    def confirm(self, prune: bool):
        self._confirm_launch(prune)

    def print_operations(self, prune: bool):
        deploy_plan = self._create_deploy_plan()
        stacks_to_skip = self._get_stacks_to_skip(deploy_plan, prune)
        self._print_skips(stacks_to_skip)
        if prune:
            pruner = self._make_pruner(self._context, self._make_plan)
            pruner.print_operations()

    def launch(self, prune: bool) -> int:
        deploy_plan = self._create_deploy_plan()
        stacks_to_skip = self._get_stacks_to_skip(deploy_plan, prune)
        stacks_to_prune = self._get_stacks_to_prune(deploy_plan, prune)

        self._exclude_stacks_from_plan(deploy_plan, *stacks_to_skip, *stacks_to_prune)
        self._validate_launch_for_missing_dependencies(deploy_plan, prune)

        code = 0
        if prune:
            code = self._prune()

        code = code or self._deploy(deploy_plan)
        return code

    def _create_deploy_plan(self) -> SceptrePlan:
        if not self._plan:
            plan = self._make_plan(self._context)
            # The plan must be resolved so we can modify launch order and items before executing it
            plan.resolve(plan.launch.__name__)
            self._plan = plan
        return self._plan

    def _get_stacks_to_skip(self, deploy_plan: SceptrePlan, prune: bool) -> List[Stack]:
        return [stack for stack in deploy_plan if stack.ignore or (stack.obsolete and not prune)]

    def _get_stacks_to_prune(self, deploy_plan: SceptrePlan, prune: bool) -> List[Stack]:
        return [stack for stack in deploy_plan if prune and stack.obsolete]

    def _exclude_stacks_from_plan(self, deployment_plan: SceptrePlan, *stacks: Stack):
        for stack in stacks:
            deployment_plan.remove_stack_from_plan(stack)

    def _validate_launch_for_missing_dependencies(self, deploy_plan: SceptrePlan, prune: bool):
        validated_stacks = set()
        skipped_dependencies = set()

        def validate_stack_dependencies(stack: Stack):
            if stack in validated_stacks:
                # In order to avoid unnecessary recursions on stacks already evaluated, we'll return
                # early if we've already evaluated the stack without issue.
                return
            if prune and stack.obsolete:
                raise DependencyDoesNotExistError(
                    f"Launch plan with --prune option depends on stack '{stack.name}' that is marked "
                    f"as obsolete. Only obsolete stacks can depend upon obsolete stacks when pruning."
                )
            for dependency in stack.dependencies:
                if dependency.ignore or dependency.obsolete:
                    skipped_dependencies.add(dependency)
                if not self._context.ignore_dependencies:
                    validate_stack_dependencies(dependency)
            validated_stacks.add(stack)

        for stack in deploy_plan:
            validate_stack_dependencies(stack)

        message = (
            "WARNING: Launch plan depends on the following ignored and/or obsolete stacks.\n"
            "  Sceptre will attempt to continue with launch, but it may fail if any Stack Configs \n"
            "  require certain resources or outputs that don't currently exist."
        )
        self._print_stacks_with_message(list(skipped_dependencies), message)

    def _print_skips(self, stacks_to_skip: List[Stack]):
        skip_message = "During launch, the following stacks will be skipped, neither created nor updated:"
        self._print_stacks_with_message(stacks_to_skip, skip_message)

    def _print_stacks_with_message(self, stacks: List[Stack], message: str):
        if not len(stacks):
            return

        message = f'* {message}\n'
        for stack in stacks:
            message += f"{Fore.YELLOW}{stack.name}{Style.RESET_ALL}\n"

        click.echo(message)

    def _print_deletions(self, stacks_to_prune: List[Stack]):
        delete_message = "During launch, the following stacks will be will be deleted, if they exist:"
        self._print_stacks_with_message(stacks_to_prune, delete_message)

    def _confirm_launch(self, prune: bool):
        operation_name = "launch"
        if prune:
            operation_name += " --prune"
        confirmation(operation_name, False, command_path=self._context.command_path)

    def _prune(self) -> int:
        pruner = self._make_pruner(self._context, self._make_plan)
        exit_code = pruner.prune()
        if exit_code != 0:
            click.echo("Stack deletion failed, so could not proceed with launch.")
        return exit_code

    def _deploy(self, deploy_plan: SceptrePlan) -> int:
        result = deploy_plan.launch()
        exit_code = stack_status_exit_code(result.values())
        return exit_code



import datetime
import json
import re
from abc import abstractmethod
from typing import TextIO, Generic, List

import cfn_flip
import yaml
from deepdiff import DeepDiff
from deepdiff.serialization import json_convertor_default

from sceptre.diffing.stack_differ import StackConfiguration, StackDiff, DiffType

deepdiff_json_defaults = {
    datetime.date: lambda x: x.isoformat(),
    StackConfiguration: lambda x: dict(x._asdict())
}


class DiffWriter(Generic[DiffType]):
    """A component responsible for taking a StackDiff and writing it in a way that is useful and
    readable. This is an abstract base class, so the abstract methods need to be implemented to
    create a DiffWriter for a given DiffType.
    """
    STAR_BAR = '*' * 80
    LINE_BAR = '-' * 80

    def __init__(self, stack_diff: StackDiff, output_stream: TextIO, output_format: str):
        """Initializes the DiffWriter

        :param stack_diff: The diff this writer will be outputting
        :param output_stream: The stream this writer should output to; Generally, this will be
            stdout
        :param output_format: Output format specified for the base Sceptre cli command; This should
            be one of "yaml", "json", or "text"
        """
        self.stack_name = stack_diff.stack_name
        self.stack_diff = stack_diff

        self.template_diff = stack_diff.template_diff
        self.config_diff = stack_diff.config_diff
        self.is_deployed = stack_diff.is_deployed

        self.output_stream = output_stream
        self.output_format = output_format

    @property
    def has_difference(self) -> bool:
        return self.has_config_difference or self.has_template_difference

    def write(self):
        """Writes the diff to the output stream."""
        self._output(self.STAR_BAR)
        if not self.has_difference:
            self._output(f"No difference to deployed stack {self.stack_name}")
            return

        self._output(f"--> Difference detected for stack {self.stack_name}!")

        if not self.is_deployed:
            self._write_new_stack_details()
            return

        self._output(self.LINE_BAR)
        self._write_config_difference()
        self._output(self.LINE_BAR)
        self._write_template_difference()

    def _write_new_stack_details(self):
        stack_config_text = self._dump_stack_config(self.stack_diff.generated_config)
        self._output(
            'This stack is not deployed yet!',
            self.LINE_BAR,
            'New Config:',
            '',
            stack_config_text,
            self.LINE_BAR,
            'New Template:',
            '',
            self.stack_diff.generated_template
        )
        return

    def _output(self, *lines: str):
        lines_with_breaks = [f'{line}\n' for line in lines]
        self.output_stream.writelines(lines_with_breaks)

    def _dump_stack_config(self, stack_config: StackConfiguration) -> str:
        stack_config_dict = dict(stack_config._asdict())
        dumped = self._dump_dict(stack_config_dict)
        return dumped

    def _dump_dict(self, dict_to_dump: dict) -> str:
        if self.output_format == "json":
            # There's not really a viable way to dump a template as "text" -> YAML is very readable
            dumper = cfn_flip.dump_json
        else:
            dumper = cfn_flip.dump_yaml

        dumped = dumper(dict_to_dump)
        return dumped

    def _write_config_difference(self):
        if not self.has_config_difference:
            self._output("No stack config difference")
            return

        diff_text = self.dump_diff(self.config_diff)
        self._output(
            f'Config difference for {self.stack_name}:',
            '',
            diff_text
        )

    def _write_template_difference(self):
        if not self.has_template_difference:
            self._output('No template difference')
            return

        diff_text = self.dump_diff(self.template_diff)
        self._output(
            f'Template difference for {self.stack_name}:',
            '',
            diff_text
        )

    @abstractmethod
    def dump_diff(self, diff: DiffType) -> str:
        """"Implement this method to write the DiffType to string"""

    @property
    @abstractmethod
    def has_config_difference(self) -> bool:
        """Implement this to indicate whether or not there is a config difference"""

    @property
    @abstractmethod
    def has_template_difference(self) -> bool:
        """Implement this to indicate whether or not there is a template difference"""


class DeepDiffWriter(DiffWriter[DeepDiff]):
    """A DiffWriter for StackDiffs where the DiffType is a DeepDiff object."""

    @property
    def has_config_difference(self) -> bool:
        return len(self.config_diff) > 0

    @property
    def has_template_difference(self) -> bool:
        return len(self.template_diff) > 0

    def dump_diff(self, diff: DeepDiff) -> str:
        as_diff_dict = diff.to_dict()
        if self.output_format == 'json':
            return json.dumps(
                as_diff_dict,
                indent=4,
                default=json_convertor_default(default_mapping=deepdiff_json_defaults)
            )

        compatible = self._make_strings_block_compatible(as_diff_dict)
        return yaml.dump(compatible, indent=4)

    def _make_strings_block_compatible(self, obj):
        """A recursive method that strips out extraneous spaces that precede line breaks.

        PyYaml disallows block styling for multiline strings if any of the lines has a space followed
        by a line break.

        DeepDiff will actually provide a difflib-style diff for multiline strings when there has
        been a value changed from one multiline string to another multiline string. However, when
        it produces that diff, every line ends with at least one space. This keeps it from being
        formatted as a block (the most useful way to display it) by PyYaml. Therefore, this function
        recurses into the deepdiff-generated data structure and strips all strings of those extraneous
        spaces that precede line breaks.

        :param obj: The DeepDiff generated diff dict (or some value this method has recursed into
            from that dict).
        :return: The object, stripped of extraneous spaces that precede line breaks.
        """
        if isinstance(obj, dict):
            return {
                key: self._make_strings_block_compatible(value)
                for key, value in obj.items()
            }
        elif isinstance(obj, list):
            return [self._make_strings_block_compatible(item) for item in obj]
        elif isinstance(obj, str):
            return re.sub('[ ]*\n', '\n', obj)
        else:
            return obj


class DiffLibWriter(DiffWriter[List[str]]):
    """A DiffWriter for StackDiffs where the DiffType is a a list of strings."""

    @property
    def has_config_difference(self) -> bool:
        return len(self.config_diff) > 0

    @property
    def has_template_difference(self) -> bool:
        return len(self.template_diff) > 0

    def dump_diff(self, diff: List[str]) -> str:
        # Difflib doesn't care about the output format since it only outputs strings. We would have
        # accounted for the output format in the differ itself rather than here.
        return '\n'.join(diff)



from os import path
from unittest.mock import sentinel
from sceptre.context import SceptreContext


class TestSceptreContext(object):

    def setup_method(self, test_method):
        self.templates_path = "templates"
        self.config_path = "config"
        self.config_file = "config.yaml"

    def test_context_with_path(self):
        self.context = SceptreContext(
            project_path="project_path/to/sceptre",
            command_path="command-path",
            user_variables=sentinel.user_variables,
            options=sentinel.options,
            output_format=sentinel.output_format,
            no_colour=sentinel.no_colour,
            ignore_dependencies=sentinel.ignore_dependencies
        )

        sentinel.project_path = "project_path/to/sceptre"
        assert self.context.project_path.replace(path.sep, '/') == sentinel.project_path

    def test_full_config_path_returns_correct_path(self):
        context = SceptreContext(
            project_path="project_path",
            command_path="command-path",
            user_variables=sentinel.user_variables,
            options=sentinel.options,
            output_format=sentinel.output_format,
            no_colour=sentinel.no_colour,
            ignore_dependencies=sentinel.ignore_dependencies
        )

        full_config_path = path.join("project_path", self.config_path)
        assert context.full_config_path() == full_config_path

    def test_full_command_path_returns_correct_path(self):
        context = SceptreContext(
            project_path="project_path",
            command_path="command",
            user_variables=sentinel.user_variables,
            options=sentinel.options,
            output_format=sentinel.output_format,
            no_colour=sentinel.no_colour,
            ignore_dependencies=sentinel.ignore_dependencies
        )
        full_command_path = path.join("project_path",
                                      self.config_path,
                                      "command")

        assert context.full_command_path() == full_command_path

    def test_full_templates_path_returns_correct_path(self):
        context = SceptreContext(
            project_path="project_path",
            command_path="command",
            user_variables=sentinel.user_variables,
            options=sentinel.options,
            output_format=sentinel.output_format,
            no_colour=sentinel.no_colour,
            ignore_dependencies=sentinel.ignore_dependencies
        )
        full_templates_path = path.join("project_path", self.templates_path)
        assert context.full_templates_path() == full_templates_path

    def test_clone__returns_full_clone_of_context(self):
        context = SceptreContext(
            project_path="project_path",
            command_path="command",
            user_variables={"user": "variables"},
            options={"hello": "there"},
            output_format=sentinel.output_format,
            no_colour=sentinel.no_colour,
            ignore_dependencies=sentinel.ignore_dependencies
        )
        clone = context.clone()
        assert clone is not context
        assert clone.project_path == context.project_path
        assert clone.command_path == context.command_path
        assert clone.user_variables == context.user_variables
        assert clone.user_variables is not context.user_variables
        assert clone.options == context.options
        assert clone.options is not context.options
        assert clone.output_format == context.output_format
        assert clone.no_colour == context.no_colour
        assert clone.ignore_dependencies == context.ignore_dependencies



import click
from click import Context

from sceptre.context import SceptreContext
from sceptre.plan.plan import SceptrePlan

from sceptre.cli.helpers import (
    catch_exceptions,
    deserialize_json_properties,
    write
)

BAD_STATUSES = [
    "DETECTION_FAILED",
    "TIMED_OUT"
]


@click.group(name="drift")
def drift_group():
    """
    Commands for calling drift detection.
    """
    pass


@drift_group.command(name="detect", short_help="Run detect stack drift on running stacks.")
@click.argument("path")
@click.pass_context
@catch_exceptions
def drift_detect(ctx: Context, path: str):
    """
    Detect stack drift and return stack drift status.

    In the event that the stack does not exist, we return
    a DetectionStatus and StackDriftStatus of STACK_DOES_NOT_EXIST.

    In the event that drift detection times out, we return
    a DetectionStatus and StackDriftStatus of TIMED_OUT.

    The timeout is set at 5 minutes, a value that cannot be configured.
    """
    context = SceptreContext(
        command_path=path,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        options=ctx.obj.get("options"),
        output_format=ctx.obj.get("output_format"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies")
    )

    plan = SceptrePlan(context)
    responses = plan.drift_detect()

    output_format = "json" if context.output_format == "json" else "yaml"

    exit_status = 0
    for stack, response in responses.items():
        status = response["DetectionStatus"]
        if status in BAD_STATUSES:
            exit_status += 1
        for key in ["Timestamp", "ResponseMetadata"]:
            response.pop(key, None)
        write({stack.external_name: deserialize_json_properties(response)}, output_format)

    exit(exit_status)


@drift_group.command(name="show", short_help="Shows stack drift on running stacks.")
@click.argument("path")
@click.option(
    "-D", "--drifted", is_flag=True, default=False, help="Filter out in sync resources."
)
@click.pass_context
@catch_exceptions
def drift_show(ctx, path, drifted):
    """
    Show stack drift on deployed stacks.

    In the event that the stack does not exist, we return
    a StackResourceDriftStatus of STACK_DOES_NOT_EXIST.

    In the event that drift detection times out, we return
    a StackResourceDriftStatus of TIMED_OUT.

    The timeout is set at 5 minutes, a value that cannot be configured.
    """
    context = SceptreContext(
        command_path=path,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        options=ctx.obj.get("options"),
        output_format=ctx.obj.get("output_format"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies")
    )

    plan = SceptrePlan(context)
    responses = plan.drift_show(drifted)

    output_format = "json" if context.output_format == "json" else "yaml"

    exit_status = 0
    for stack, (status, response) in responses.items():
        if status in BAD_STATUSES:
            exit_status += 1
        write({stack.external_name: deserialize_json_properties(response)}, output_format)

    exit(exit_status)



# -*- coding: utf-8 -*-

import pytest
from unittest.mock import MagicMock, patch, sentinel

from sceptre.exceptions import DependencyStackMissingOutputError
from sceptre.exceptions import StackDoesNotExistError
from botocore.exceptions import ClientError

from sceptre.connection_manager import ConnectionManager
from sceptre.resolvers.stack_output import \
    StackOutput, StackOutputExternal, StackOutputBase
from sceptre.stack import Stack


class TestStackOutputResolver(object):

    @patch(
        "sceptre.resolvers.stack_output.StackOutput._get_output_value"
    )
    def test_resolver(self, mock_get_output_value):
        stack = MagicMock(spec=Stack)
        stack.dependencies = []
        stack.project_code = "project-code"
        stack._connection_manager = MagicMock(spec=ConnectionManager)

        dependency = MagicMock()
        dependency.project_code = "meh"
        dependency.name = "account/dev/vpc"
        dependency.profile = "dependency_profile"
        dependency.region = "dependency_region"
        dependency.iam_role = "dependency_iam_role"

        mock_get_output_value.return_value = "output_value"

        stack_output_resolver = StackOutput(
            "account/dev/vpc.yaml::VpcId", stack
        )

        stack_output_resolver.setup()
        assert stack.dependencies == ["account/dev/vpc.yaml"]

        stack.dependencies = [dependency]
        result = stack_output_resolver.resolve()
        assert result == "output_value"
        mock_get_output_value.assert_called_once_with(
            "meh-account-dev-vpc", "VpcId",
            profile="dependency_profile", region="dependency_region",
            iam_role="dependency_iam_role"
        )

    @patch(
        "sceptre.resolvers.stack_output.StackOutput._get_output_value"
    )
    def test_resolver_with_existing_dependencies(self, mock_get_output_value):
        stack = MagicMock(spec=Stack)
        stack.dependencies = ["existing"]
        stack.project_code = "project-code"
        stack._connection_manager = MagicMock(spec=ConnectionManager)

        dependency = MagicMock()
        dependency.project_code = "meh"
        dependency.name = "account/dev/vpc"
        dependency.profile = "dependency_profile"
        dependency.region = "dependency_region"
        dependency.iam_role = "dependency_iam_role"

        mock_get_output_value.return_value = "output_value"

        stack_output_resolver = StackOutput(
            "account/dev/vpc.yaml::VpcId", stack
        )

        stack_output_resolver.setup()
        assert stack.dependencies == ["existing", "account/dev/vpc.yaml"]

        stack.dependencies = [MagicMock(), dependency]
        result = stack_output_resolver.resolve()
        assert result == "output_value"
        mock_get_output_value.assert_called_once_with(
            "meh-account-dev-vpc", "VpcId",
            profile="dependency_profile", region="dependency_region",
            iam_role="dependency_iam_role"
        )

    @patch(
        "sceptre.resolvers.stack_output.StackOutput._get_output_value"
    )
    def test_resolve_with_implicit_stack_reference(
        self, mock_get_output_value
    ):
        stack = MagicMock(spec=Stack)
        stack.dependencies = []
        stack.project_code = "project-code"
        stack.name = "account/dev/stack"
        stack._connection_manager = MagicMock(spec=ConnectionManager)

        dependency = MagicMock()
        dependency.project_code = "meh"
        dependency.name = "account/dev/vpc"
        dependency.profile = "dependency_profile"
        dependency.region = "dependency_region"
        dependency.iam_role = "dependency_iam_role"

        mock_get_output_value.return_value = "output_value"

        stack_output_resolver = StackOutput("account/dev/vpc.yaml::VpcId", stack)

        stack_output_resolver.setup()
        assert stack.dependencies == ["account/dev/vpc.yaml"]

        stack.dependencies = [dependency]
        result = stack_output_resolver.resolve()
        assert result == "output_value"
        mock_get_output_value.assert_called_once_with(
            "meh-account-dev-vpc", "VpcId",
            profile="dependency_profile", region="dependency_region",
            iam_role="dependency_iam_role"
        )

    @patch(
        "sceptre.resolvers.stack_output.StackOutput._get_output_value"
    )
    def test_resolve_with_implicit_stack_reference_top_level(
        self, mock_get_output_value
    ):
        stack = MagicMock(spec=Stack)
        stack.dependencies = []
        stack.project_code = "project-code"
        stack.name = "stack"
        stack._connection_manager = MagicMock(spec=ConnectionManager)

        dependency = MagicMock()
        dependency.project_code = "meh"
        dependency.name = "vpc"
        dependency.profile = "dependency_profile"
        dependency.region = "dependency_region"
        dependency.iam_role = "dependency_iam_role"

        mock_get_output_value.return_value = "output_value"

        stack_output_resolver = StackOutput("vpc.yaml::VpcId", stack)

        stack_output_resolver.setup()
        assert stack.dependencies == ["vpc.yaml"]

        stack.dependencies = [dependency]
        result = stack_output_resolver.resolve()
        assert result == "output_value"
        mock_get_output_value.assert_called_once_with(
            "meh-vpc", "VpcId",
            profile="dependency_profile", region="dependency_region",
            iam_role="dependency_iam_role"
        )


class TestStackOutputExternalResolver(object):

    @patch(
        "sceptre.resolvers.stack_output.StackOutputExternal._get_output_value"
    )
    def test_resolve(self, mock_get_output_value):
        stack = MagicMock(spec=Stack)
        stack.dependencies = []
        stack._connection_manager = MagicMock(spec=ConnectionManager)
        stack_output_external_resolver = StackOutputExternal(
            "another/account-vpc::VpcId", stack
        )
        mock_get_output_value.return_value = "output_value"
        stack_output_external_resolver.resolve()
        mock_get_output_value.assert_called_once_with(
            "another/account-vpc", "VpcId", None, None, None
        )
        assert stack.dependencies == []

    @patch(
        "sceptre.resolvers.stack_output.StackOutputExternal._get_output_value"
    )
    def test_resolve_with_args(self, mock_get_output_value):
        stack = MagicMock(spec=Stack)
        stack.dependencies = []
        stack._connection_manager = MagicMock(spec=ConnectionManager)
        stack_output_external_resolver = StackOutputExternal(
            "another/account-vpc::VpcId region::profile::iam_role", stack
        )
        mock_get_output_value.return_value = "output_value"
        stack_output_external_resolver.resolve()
        mock_get_output_value.assert_called_once_with(
            "another/account-vpc", "VpcId", "region", "profile", "iam_role"
        )
        assert stack.dependencies == []


class MockStackOutputBase(StackOutputBase):
    """
    MockBaseResolver inherits from the abstract base class
    StackOutputBaseResolver, and implements the abstract methods. It is used
    to allow testing on StackOutputBaseResolver, which is not otherwise
    instantiable.
    """

    def __init__(self, *args, **kwargs):
        super(MockStackOutputBase, self).__init__(*args, **kwargs)

    def resolve(self):
        pass


class TestStackOutputBaseResolver(object):

    def setup_method(self, test_method):
        self.stack = MagicMock(spec=Stack)
        self.stack._connection_manager = MagicMock(
            spec=ConnectionManager
        )
        self.base_stack_output_resolver = MockStackOutputBase(
            None, self.stack
        )

    @patch(
        "sceptre.resolvers.stack_output.StackOutputBase._get_stack_outputs"
    )
    def test_get_output_value_with_valid_key(self, mock_get_stack_outputs):
        mock_get_stack_outputs.return_value = {"key": "value"}

        response = self.base_stack_output_resolver._get_output_value(
            sentinel.stack_name, "key"
        )

        assert response == "value"

    @patch(
        "sceptre.resolvers.stack_output.StackOutputBase._get_stack_outputs"
    )
    def test_get_output_value_with_invalid_key(self, mock_get_stack_outputs):
        mock_get_stack_outputs.return_value = {"key": "value"}

        with pytest.raises(DependencyStackMissingOutputError):
            self.base_stack_output_resolver._get_output_value(
                sentinel.stack_name, "invalid_key"
            )

    def test_get_stack_outputs_with_valid_stack(self):
        self.stack.connection_manager.call.return_value = {
            "Stacks": [{
                "Outputs": [
                    {
                        "OutputKey": "key_1",
                        "OutputValue": "value_1",
                        "Description": "description_1"
                    },
                    {
                        "OutputKey": "key_2",
                        "OutputValue": "value_2",
                        "Description": "description_2"
                    }
                ]
            }]
        }

        response = self.base_stack_output_resolver._get_stack_outputs(
            sentinel.stack_name
        )

        assert response == {
            "key_1": "value_1",
            "key_2": "value_2"
        }

    def test_get_stack_outputs_with_valid_stack_without_outputs(self):
        self.stack.connection_manager.call.return_value = {
            "Stacks": [{}]
        }

        response = self.base_stack_output_resolver._get_stack_outputs(
            sentinel.stack_name
        )
        assert response == {}

    def test_get_stack_outputs_with_unlaunched_stack(self):
        self.stack.connection_manager.call.side_effect = ClientError(
            {
                "Error": {
                    "Code": "404",
                    "Message": "stack does not exist"
                }
            },
            sentinel.operation
        )

        with pytest.raises(StackDoesNotExistError):
            self.base_stack_output_resolver._get_stack_outputs(
                sentinel.stack_name
            )

    def test_get_stack_outputs_with_unkown_boto_error(self):
        self.stack.connection_manager.call.side_effect = ClientError(
            {
                "Error": {
                    "Code": "500",
                    "Message": "Boom!"
                }
            },
            sentinel.operation
        )

        with pytest.raises(ClientError):
            self.base_stack_output_resolver._get_stack_outputs(
                sentinel.stack_name
            )



import os
import sys
import pytest
import yaml

import sceptre.template_handlers.helper as helper
from sceptre.exceptions import TemplateNotFoundError
from unittest.mock import patch


@pytest.mark.parametrize("filename,sceptre_user_data,expected", [
    (
        "vpc.j2",
        {"vpc_id": "10.0.0.0/16"},
        """Resources:
  VPC:
    Type: AWS::EC2::VPC
    Properties:
      CidrBlock: 10.0.0.0/16
Outputs:
  VpcId:
    Value:
      Ref: VPC"""
    ),
    (
        "vpc.yaml.j2",
        {"vpc_id": "10.0.0.0/16"},
        """Resources:
  VPC:
    Type: AWS::EC2::VPC
    Properties:
      CidrBlock: 10.0.0.0/16
Outputs:
  VpcId:
    Value:
      Ref: VPC"""
    ),
    (
        "sg.j2",
        [
            {"name": "sg_a", "inbound_ip": "10.0.0.0"},
            {"name": "sg_b", "inbound_ip": "10.0.0.1"}
        ],
        """Resources:
    sg_a:
        Type: "AWS::EC2::SecurityGroup"
        Properties:
            InboundIp: 10.0.0.0
    sg_b:
        Type: "AWS::EC2::SecurityGroup"
        Properties:
            InboundIp: 10.0.0.1
"""
    )
])
@patch("pathlib.Path.exists")
def test_render_jinja_template(mock_pathlib,
                               filename,
                               sceptre_user_data,
                               expected):
    mock_pathlib.return_value = True
    jinja_template_path = os.path.join(
        os.getcwd(),
        "tests/fixtures/templates",
        filename
    )
    result = helper.render_jinja_template(path=jinja_template_path,
                                          jinja_vars={"sceptre_user_data": sceptre_user_data},
                                          j2_environment={})
    expected_yaml = yaml.safe_load(expected)
    result_yaml = yaml.safe_load(result)
    assert expected_yaml == result_yaml


@pytest.mark.skipif(sys.version_info < (3, 8), reason="requires Python >= 3.8")
@pytest.mark.parametrize("j2_environment,expected_keys", [
    ({}, ["autoescape", "loader", "undefined"]),
    ({"lstrip_blocks": True},
     ["autoescape", "loader", "undefined", "lstrip_blocks"]),
    ({"lstrip_blocks": True, "extensions": ["test-ext"]},
     ["autoescape", "loader", "undefined", "lstrip_blocks", "extensions"])
])
@patch("sceptre.template_handlers.helper.Environment")
@patch("pathlib.Path.exists")
def test_render_jinja_template_j2_environment_config(mock_pathlib,
                                                     mock_environment,
                                                     j2_environment,
                                                     expected_keys):
    mock_pathlib.return_value = True
    filename = "vpc.j2"
    sceptre_user_data = {"vpc_id": "10.0.0.0/16"}
    jinja_template_path = os.path.join(
        os.getcwd(),
        "tests/fixtures/templates",
        filename
    )
    _ = helper.render_jinja_template(path=jinja_template_path,
                                     jinja_vars={"sceptre_user_data": sceptre_user_data},
                                     j2_environment=j2_environment)
    assert list(mock_environment.call_args.kwargs) == expected_keys


def test_render_jinja_template_non_existing_file():
    jinja_template_path = os.path.join("/ref/to/nowhere/boom.j2")
    with pytest.raises(TemplateNotFoundError):
        helper.render_jinja_template(path=jinja_template_path,
                                     jinja_vars={"sceptre_user_data": {}},
                                     j2_environment={})



# -*- coding: utf-8 -*-
import pathlib
import requests
import tempfile
import sceptre.template_handlers.helper as helper

from requests.adapters import HTTPAdapter
from requests.exceptions import InvalidURL, ConnectTimeout
from requests.packages.urllib3.util.retry import Retry
from sceptre.exceptions import UnsupportedTemplateFileTypeError
from sceptre.template_handlers import TemplateHandler
from urllib.parse import urlparse

HANDLER_OPTION_KEY = "http_template_handler"
HANDLER_RETRIES_OPTION_PARAM = "retries"
DEFAULT_RETRIES_OPTION = 5
HANDLER_TIMEOUT_OPTION_PARAM = "timeout"
DEFAULT_TIMEOUT_OPTION = 5


class Http(TemplateHandler):
    """
    Template handler that can resolve templates from the web.  Standard CFN templates
    with extension (.json, .yaml, .template) are deployed directly from memory
    while references to jinja (.j2) and python (.py) templates are downloaded,
    transformed into CFN templates then deployed to AWS.
    """
    def __init__(self, *args, **kwargs):
        super(Http, self).__init__(*args, **kwargs)

    def schema(self):
        return {
            "type": "object",
            "properties": {
                "url": {"type": "string"}
            },
            "required": ["url"]
        }

    def handle(self):
        """
        handle template from web
        """
        url = self.arguments["url"]
        path = pathlib.Path(urlparse(url).path)

        if path.suffix not in self.supported_template_extensions:
            raise UnsupportedTemplateFileTypeError(
                "Template has file extension %s. Only %s are supported.",
                path.suffix, ",".join(self.supported_template_extensions)
            )

        retries = self._get_handler_option(HANDLER_RETRIES_OPTION_PARAM, DEFAULT_RETRIES_OPTION)
        timeout = self._get_handler_option(HANDLER_TIMEOUT_OPTION_PARAM, DEFAULT_TIMEOUT_OPTION)
        try:
            template = self._get_template(url, retries=retries, timeout=timeout)
            if path.suffix in self.jinja_template_extensions + self.python_template_extensions:
                file = tempfile.NamedTemporaryFile(prefix=path.stem)
                self.logger.debug("Template file saved to: %s", file.name)
                with file as f:
                    f.write(template)
                    f.seek(0)
                    f.read()
                    if path.suffix in self.jinja_template_extensions:
                        template = helper.render_jinja_template(f.name,
                                                                {"sceptre_user_data": self.sceptre_user_data},
                                                                self.stack_group_config.get("j2_environment", {}))
                    elif path.suffix in self.python_template_extensions:
                        template = helper.call_sceptre_handler(
                            f.name,
                            self.sceptre_user_data
                        )

        except Exception as e:
            helper.print_template_traceback(path)
            raise e

        return template

    def _get_template(self, url, retries, timeout):
        """
        Get template from the web
        :param url: The url to the template
        :type: str
        :param retries: The number of retry attempts.
        :rtype: int
        :param timeout: The timeout for the session in seconds.
        :rtype: int
        """
        self.logger.debug("Downloading file from: %s", url)
        session = self._get_retry_session(retries=retries)
        try:
            response = session.get(url, timeout=timeout)
            return response.content
        except (InvalidURL, ConnectTimeout) as e:
            raise e

    def _get_retry_session(self,
                           retries,
                           backoff_factor=0.3,
                           status_forcelist=(429, 500, 502, 503, 504),
                           session=None):
        """
        Get a request session with retries.  Retry options are explained in the request libraries
        https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html#module-urllib3.util.retry
        """
        session = session or requests.Session()
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
        )
        adapter = HTTPAdapter(max_retries=retry)
        session.mount('http://', adapter)
        session.mount('https://', adapter)
        return session

    def _get_handler_option(self, name, default):
        """
        Get the template handler options
        :param url: The option name
        :type: str
        :param default: The default value if option is not set.
        :rtype: int
        """
        if HANDLER_OPTION_KEY in self.stack_group_config:
            option = self.stack_group_config.get(HANDLER_OPTION_KEY)
            if name in option:
                return option.get(name)

        return default
